cas_cache_clear ();;
cas_net := basic_net ();;

(* Treatment of assumptions *)
cas_assume_pattern `x` REAL_ARITH;
cas_assume_pattern `x` (conv_or_not_conv ARITH_CONV);
cas_pattern `x` boolassume_conv;
cas_pattern `assume x` assume_conv;
cas_pattern `assume_type x` assume_conv;
cas_pattern `x assuming y` (assuming_conv cas_conv);

(* Before quotients get reduced *)
cas_pattern `&(NUMERAL x) / &(NUMERAL y)` decimal_conv;;

(* If the nets would be outside, we could merge them, since they are not it's copied *)
(* NUM_REDUCE *)
cas_pattern `SUC(NUMERAL n)` NUM_SUC_CONV;
cas_pattern `PRE(NUMERAL n)` NUM_PRE_CONV;
cas_pattern `FACT(NUMERAL n)` NUM_FACT_CONV;
cas_pattern `NUMERAL m < NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m <= NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m > NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m >= NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m = NUMERAL n` NUM_REL_CONV;
cas_pattern `EVEN(NUMERAL n)` NUM_EVEN_CONV;
cas_pattern `ODD(NUMERAL n)` NUM_ODD_CONV;
cas_pattern `NUMERAL m + NUMERAL n` NUM_ADD_CONV;
cas_pattern `NUMERAL m - NUMERAL n` NUM_SUB_CONV;
cas_pattern `NUMERAL m * NUMERAL n` NUM_MULT_CONV;
cas_pattern `(NUMERAL m) EXP (NUMERAL n)` NUM_EXP_CONV;
cas_pattern `(NUMERAL m) DIV (NUMERAL n)` NUM_DIV_CONV;
cas_pattern `(NUMERAL m) MOD (NUMERAL n)` NUM_MOD_CONV;

prioritize_real ();

(* Are these necessary, or are the reational versions more powerful? *)
(* REAL_INT_RED *)
cas_pattern `x <= y` REAL_INT_LE_CONV;
cas_pattern `x < y` REAL_INT_LT_CONV;
cas_pattern `x >= y` REAL_INT_GE_CONV;
cas_pattern `x > y` REAL_INT_GT_CONV;
cas_pattern `x:real = y` REAL_INT_EQ_CONV;
cas_pattern `--x` (CHANGED_CONV REAL_INT_NEG_CONV);
cas_pattern `abs(x)` REAL_INT_ABS_CONV;
cas_pattern `x + y` REAL_INT_ADD_CONV;
cas_pattern `x - y` REAL_INT_SUB_CONV;
cas_pattern `x * y` REAL_INT_MUL_CONV;
cas_pattern `x pow n` REAL_INT_POW_CONV;

(* REAL_RAT_RED *)
cas_pattern `x <= y` REAL_RAT_LE_CONV; ;
cas_pattern `x < y` REAL_RAT_LT_CONV;
cas_pattern `x >= y` REAL_RAT_GE_CONV;
cas_pattern `x > y` REAL_RAT_GT_CONV;
cas_pattern `x:real = y` REAL_RAT_EQ_CONV;
cas_pattern `--x` (CHANGED_CONV REAL_RAT_NEG_CONV);
cas_pattern `abs(x)` REAL_RAT_ABS_CONV;
cas_pattern `inv(x)` REAL_RAT_INV_CONV;
cas_pattern `x + y` REAL_RAT_ADD_CONV;
cas_pattern `x - y` REAL_RAT_SUB_CONV;
cas_pattern `x * y` REAL_RAT_MUL_CONV;
cas_pattern `x / y` REAL_RAT_DIV_CONV;
cas_pattern `x pow n` REAL_RAT_POW_CONV;;

(* Complete for ints, since normalization knows only these *)
List.iter (fun x -> cas_pattern x cas_int_conv)
  [`int_add x y`; `int_sub x y`; `int_mul x y`; `int_lt x y`;
   `int_le x y`; `int_gt x y`; `int_ge x y`; `int_neg x`;
   `int_pow x y`; `int_abs x`; `int_max x y`; `int_min x y`;
   `mod_int x y z`]
;

List.iter (fun x -> cas_pattern x REAL_POLY_CONV)
  [`real_add x y`; `real_sub x y`; `real_mul x y`; `real_div x y`;
   `real_lt x y`; `real_le x y`; `real_gt x y`; `real_ge x y`;
   `real_neg x`; `real_pow x y`; `real_inv x`; `real_abs x`;
   `real_max x y`; `real_min x y`; `mod_real x y z`]
;

(* !!! May be incomplete !!! *)
List.iter (fun x -> cas_pattern x NUM_NORMALIZE_CONV)
  [`(x:num) + y`; `(x:num) * y`; `(x:num) - y`]
;

(* Every is added with a "not" variant *)
List.iter (fun x -> cas_pattern x (conv_or_not_conv ARITH_CONV))
  [`(x:num) > y`; `(x:num) >= y`; `(x:num) < y`; `(x:num) <= y`; `(x:num) = y`]
;


cas_pattern `inv(x) * inv(y)`  (REWRITE_CONV[GSYM REAL_INV_MUL]); (* Multiply out real inverses *)
cas_pattern `(x:num) <= y`  (REWRITE_CONV[LE_REFL]);
cas_pattern `diff f`  diff_conv;
cas_pattern `dint range f`  int_conv;
cas_pattern `sum range f`  sum_conv;
cas_pattern `N x p`  evalf_conv;
cas_pattern `binom (n,m)`  binom_conv;
cas_pattern `n divides m`   DIVIDES_CONV;
cas_pattern `prime n`   prime_conv;;

(* For solving: `((x:num) + y) - (x + z) >= y - z` *)
List.iter cas_eq_thm [GT; GE; SUB_REFL; SUB_PRESUC; SUB_ADD_LCANCEL; SUB_ADD_RCANCEL; SUB_0];
List.iter cas_eq_thm cas_vector_thms;


(* COMPLEX *)
cas_eq_thm RE;
cas_eq_thm IM;
cas_eq_thm CX_DEF;
cas_eq_thm ii;
cas_eq_thm complex_sub;
cas_eq_thm COMPLEX;
cas_pattern `complex(x,y) + complex(z,v)` (REWRITE_CONV[complex_add]);
cas_pattern `complex(x,y) * complex(z,v)` (REWRITE_CONV[complex_mul]);
cas_pattern `inv (complex(z,v))` (REWRITE_CONV[complex_inv]);
cas_pattern `-- (complex(z,v))` (REWRITE_CONV[complex_neg]);
cas_pattern `complex(&0, &1) pow n`  COMPLEX_II_POW_CONV; (* My II_POW_CONV *)
List.iter (fun x -> cas_pattern x complex_poly_conv)
  [`complex_add x y`; `complex_sub x y`; `complex_mul x y`; `complex_div x y`;
   `complex_neg x`; `complex_pow x y`; `complex_inv x`]
;
cas_pattern `norm(complex(x,y))` (REWRITE_CONV[complex_norm]);
cas_eq_thm COMPLEX_ABS_NORM;
cas_pattern `cnj(complex(x,y))` (REWRITE_CONV[cnj]);
cas_eq_thm CNJ_CNJ;

cas_eq_thm REAL_MUL_LID;
cas_eq_thm REAL_MUL_RID;

cas_eq_thm ETA_AX;


cas_eq_thm POW_2_SQRT_ABS;
cas_eq_thm SQRT_0;
cas_eq_thm SQRT_1;

cas_impl_thm abs_thm;
cas_impl_thm REAL_MUL_LINV;
cas_impl_thm REAL_MUL_RINV;;
cas_impl_thm (prove(`x < &0 ==> abs(x) = --x`, REAL_ARITH_TAC));;

cas_eq_thm COMPLEX_EQ;;


cas_eq_thm (injectivity "option");;
cas_pattern `NONE = SOME (x:A)` (REWRITE_CONV [distinctness "option"]);;
cas_pattern `SOME (x:A) = NONE` (REWRITE_CONV [distinctness "option"]);;

let SELECT_LEMMA = prove
   (`!x:A. (@y. x = y) = x`,
    GEN_TAC THEN GEN_REWRITE_TAC (LAND_CONV o BINDER_CONV) [EQ_SYM_EQ] THEN
    MATCH_ACCEPT_TAC SELECT_REFL);;

cas_eq_thm SELECT_LEMMA;;

let option1 = prove (`(NONE = SOME (x:A)) <=> F`, REWRITE_TAC[distinctness "option"]);;

let o1 = prove (`(if P then (SOME x) else NONE) = NONE <=> ~P`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"]);;
let o2 = prove (`(if P then NONE else (SOME x)) = NONE <=> P`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"]);;
let o3 = prove (`(if P then (SOME x) else NONE) = SOME y <=> P /\ x = y`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"; injectivity "option"]);;
let o4 = prove (`(if P then NONE else (SOME x)) = SOME y <=> ~P /\ x = y`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"; injectivity "option"]);;

List.iter cas_eq_thm [o1; o2; o3; o4];;
