cas_cache_clear ();;
cas_net := basic_net ();;

(* Treatment of assumptions *)
cas_assume_pattern `x` REAL_ARITH;
cas_assume_pattern `x` (conv_or_not_conv ARITH_CONV);
cas_pattern `x` boolassume_conv;
cas_pattern `assume x` assume_conv;
cas_pattern `assume_type x` assume_conv;
cas_pattern `x assuming y` (assuming_conv cas_conv);

(* Before quotients get reduced *)
cas_pattern `&(NUMERAL x) / &(NUMERAL y)` decimal_conv;;

(* If the nets would be outside, we could merge them, since they are not it's copied *)
(* NUM_REDUCE *)
cas_pattern `SUC(NUMERAL n)` NUM_SUC_CONV;
cas_pattern `PRE(NUMERAL n)` NUM_PRE_CONV;
cas_pattern `FACT(NUMERAL n)` NUM_FACT_CONV;
cas_pattern `NUMERAL m < NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m <= NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m > NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m >= NUMERAL n` NUM_REL_CONV;
cas_pattern `NUMERAL m = NUMERAL n` NUM_REL_CONV;
cas_pattern `EVEN(NUMERAL n)` NUM_EVEN_CONV;
cas_pattern `ODD(NUMERAL n)` NUM_ODD_CONV;
cas_pattern `NUMERAL m + NUMERAL n` NUM_ADD_CONV;
cas_pattern `NUMERAL m - NUMERAL n` NUM_SUB_CONV;
cas_pattern `NUMERAL m * NUMERAL n` NUM_MULT_CONV;
cas_pattern `(NUMERAL m) EXP (NUMERAL n)` NUM_EXP_CONV;
cas_pattern `(NUMERAL m) DIV (NUMERAL n)` NUM_DIV_CONV;
cas_pattern `(NUMERAL m) MOD (NUMERAL n)` NUM_MOD_CONV;

prioritize_real ();

(* Are these necessary, or are the reational versions more powerful? *)
(* REAL_INT_RED *)
cas_pattern `x <= y` REAL_INT_LE_CONV;
cas_pattern `x < y` REAL_INT_LT_CONV;
cas_pattern `x >= y` REAL_INT_GE_CONV;
cas_pattern `x > y` REAL_INT_GT_CONV;
cas_pattern `x:real = y` REAL_INT_EQ_CONV;
cas_pattern `--x` (CHANGED_CONV REAL_INT_NEG_CONV);
cas_pattern `abs(x)` REAL_INT_ABS_CONV;
cas_pattern `x + y` REAL_INT_ADD_CONV;
cas_pattern `x - y` REAL_INT_SUB_CONV;
cas_pattern `x * y` REAL_INT_MUL_CONV;
cas_pattern `x pow n` REAL_INT_POW_CONV;

(* REAL_RAT_RED *)
cas_pattern `x <= y` REAL_RAT_LE_CONV; ;
cas_pattern `x < y` REAL_RAT_LT_CONV;
cas_pattern `x >= y` REAL_RAT_GE_CONV;
cas_pattern `x > y` REAL_RAT_GT_CONV;
cas_pattern `x:real = y` REAL_RAT_EQ_CONV;
cas_pattern `--x` (CHANGED_CONV REAL_RAT_NEG_CONV);
cas_pattern `abs(x)` REAL_RAT_ABS_CONV;
cas_pattern `inv(x)` REAL_RAT_INV_CONV;
cas_pattern `x + y` REAL_RAT_ADD_CONV;
cas_pattern `x - y` REAL_RAT_SUB_CONV;
cas_pattern `x * y` REAL_RAT_MUL_CONV;
cas_pattern `x / y` REAL_RAT_DIV_CONV;
cas_pattern `x pow n` REAL_RAT_POW_CONV;;

(* Complete for ints, since normalization knows only these *)
List.iter (fun x -> cas_pattern x cas_int_conv)
  [`int_add x y`; `int_sub x y`; `int_mul x y`; `int_lt x y`;
   `int_le x y`; `int_gt x y`; `int_ge x y`; `int_neg x`;
   `int_pow x y`; `int_abs x`; `int_max x y`; `int_min x y`;
   `mod_int x y z`]
;

List.iter (fun x -> cas_pattern x REAL_POLY_CONV)
  [`real_add x y`; `real_sub x y`; `real_mul x y`; `real_div x y`;
   `real_lt x y`; `real_le x y`; `real_gt x y`; `real_ge x y`;
   `real_neg x`; `real_pow x y`; `real_inv x`; `real_abs x`;
   `real_max x y`; `real_min x y`; `mod_real x y z`]
;

(* !!! May be incomplete !!! *)
List.iter (fun x -> cas_pattern x NUM_NORMALIZE_CONV)
  [`(x:num) + y`; `(x:num) * y`; `(x:num) - y`]
;

(* Every is added with a "not" variant *)
List.iter (fun x -> cas_pattern x (conv_or_not_conv ARITH_CONV))
  [`(x:num) > y`; `(x:num) >= y`; `(x:num) < y`; `(x:num) <= y`; `(x:num) = y`]
;


cas_pattern `inv(x) * inv(y)`  (REWRITE_CONV[GSYM REAL_INV_MUL]); (* Multiply out real inverses *)
cas_pattern `(x:num) <= y`  (REWRITE_CONV[LE_REFL]);
cas_pattern `diff f`  diff_conv;
cas_pattern `dint range f`  int_conv;
cas_pattern `sum range f`  sum_conv;
cas_pattern `N x p`  evalf_conv;
cas_pattern `binom (n,m)`  binom_conv;
cas_pattern `n divides m`   DIVIDES_CONV;
cas_pattern `prime n`   prime_conv;;

(* For solving: `((x:num) + y) - (x + z) >= y - z` *)
List.iter cas_eq_thm [GT; GE; SUB_REFL; SUB_PRESUC; SUB_ADD_LCANCEL; SUB_ADD_RCANCEL; SUB_0];
List.iter cas_eq_thm cas_vector_thms;


(* COMPLEX *)
cas_eq_thm RE;
cas_eq_thm IM;
cas_eq_thm CX_DEF;
cas_eq_thm ii;
cas_eq_thm complex_sub;
cas_eq_thm COMPLEX;
cas_pattern `complex(x,y) + complex(z,v)` (REWRITE_CONV[complex_add]);
cas_pattern `complex(x,y) * complex(z,v)` (REWRITE_CONV[complex_mul]);
cas_pattern `inv (complex(z,v))` (REWRITE_CONV[complex_inv]);
cas_pattern `-- (complex(z,v))` (REWRITE_CONV[complex_neg]);
cas_pattern `complex(&0, &1) pow n`  COMPLEX_II_POW_CONV; (* My II_POW_CONV *)
List.iter (fun x -> cas_pattern x complex_poly_conv)
  [`complex_add x y`; `complex_sub x y`; `complex_mul x y`; `complex_div x y`;
   `complex_neg x`; `complex_pow x y`; `complex_inv x`]
;
cas_pattern `norm(complex(x,y))` (REWRITE_CONV[complex_norm]);
cas_eq_thm COMPLEX_ABS_NORM;
cas_pattern `cnj(complex(x,y))` (REWRITE_CONV[cnj]);
cas_eq_thm CNJ_CNJ;

cas_eq_thm REAL_MUL_LID;
cas_eq_thm REAL_MUL_RID;

cas_eq_thm ETA_AX;


cas_eq_thm POW_2_SQRT_ABS;
cas_eq_thm SQRT_0;
cas_eq_thm SQRT_1;

cas_impl_thm abs_thm;
cas_impl_thm REAL_MUL_LINV;
cas_impl_thm REAL_MUL_RINV;;
cas_impl_thm (prove(`x < &0 ==> abs(x) = --x`, REAL_ARITH_TAC));;

cas_eq_thm COMPLEX_EQ;;


cas_eq_thm (injectivity "option");;
cas_pattern `NONE = SOME (x:A)` (REWRITE_CONV [distinctness "option"]);;
cas_pattern `SOME (x:A) = NONE` (REWRITE_CONV [distinctness "option"]);;

let SELECT_LEMMA = prove
   (`!x:A. (@y. x = y) = x`,
    GEN_TAC THEN GEN_REWRITE_TAC (LAND_CONV o BINDER_CONV) [EQ_SYM_EQ] THEN
    MATCH_ACCEPT_TAC SELECT_REFL);;

cas_eq_thm SELECT_LEMMA;;




prioritize_real ();;
let partial_def = new_definition `!(f:A^N->A) (d:A^N->bool). partial(f, d) = (f, d)`;;
let partial_apply1 = define
  `(papp1 ((f:A^1->A), (d:A^1->bool)) NONE = NONE) /\
   (papp1 (f, d) (SOME x) = if (d (lambda i.x)) then (SOME (f (lambda i.x))) else NONE)`;;
let partial_apply2 = define
  `(papp2 ((f:A^2->A), (d:A^2->bool)) (SOME x) (SOME y) =
     if (d (lambda i.if i = 1 then x else y)) then
       (SOME (f (lambda i.if i = 1 then x else y))) else NONE) /\
   (papp2 (f, d) NONE v = NONE) /\
   (papp2 (f, d) v NONE = NONE)`;;
let partial_binop = new_definition `pbinop (f:A->A->A) x y = papp2 ((\x:A^2. (f:A->A->A) (x$1) (x$2)),(\x:A^2.T)) x y`;;
let partial_add = new_definition `padd = pbinop real_add`;;
let partial_sub = new_definition `psub = pbinop real_sub`;;
let partial_mul = new_definition `pmul = pbinop real_mul`;;
let partial_inv = new_definition `pinv = papp1 (partial ((\x:real^1. inv (x$1)), \x:real^1. ~((x$1) = &0)))`;;
let partial_div = new_definition `pdiv x y = pmul x (pinv y)`;;
let partial_num = new_definition `pnum n = SOME (&n)`;;

let partial_diff_proto = new_definition `pdiff_proto (f:real^1->real, d:real^1->bool) = 
 ( (\x:real^1. @v. ((\x. f (lambda i. x)) diffl v) (x$1)) , (\x:real^1. d x /\ ?v. ((\x. f (lambda i. x)) diffl v) (x$1)) )`;;

REWRITE_RULE [partial_apply1] 
  (MK_COMB (
    (MK_COMB (REFL `papp1:(real^1->real)#(real^1->bool)->(real)option->(real)option`, SPEC_ALL partial_diff_proto)),
     REFL `(SOME z):(real)option`)
  )
;;

let partial_unapply1 = new_definition 
  `punapp1 f = (\x:real^1. if ~(f (SOME (x$1)) = NONE) then @v. (f(SOME (x$1))) = (SOME v) else &0), (\x:real^1. ?v. (f (SOME (x$1))) = (SOME v))`;;


let partial_unapply2 = new_definition
  `punapp2 f = ((\x:real^1. @v:real. (f(SOME (x$1))) = (SOME v)), (\x:real^1. ?v. (f (SOME (x$1))) = (SOME v)))`
;;

let partial_unapply3 = new_definition
  `punapp3 f = @g. papp1 g = f`
;;

(*let partial_unapply3_fact = prove (
  `punapp3 (papp1 f) = f`, REWRITE_TAC[partial_unapply3]
);;*)

let partial_diff = new_definition `pdiff f = papp1 (pdiff_proto (punapp1 f))`;;

let lambda_eta = REWRITE_RULE [ARITH_RULE `1 <= 1`; DIMINDEX_GE_1] (SPEC `1` LAMBDA_BETA);;
let lambda_eta2 = prove (`(lambda i. x$1) = (x:A^1)`, 
  SIMP_TAC [CART_EQ;DIMINDEX_1;ARITH;LAMBDA_BETA;ARITH_RULE `1 <= i /\ i <= 1 <=> i = 1`]
);;

let partial_unapply1_fact = prove (
  `punapp1 (papp1 (f, d)) = ((\y. if d y then f y else &0), \x. d (x:real^1))`,
  REWRITE_TAC[partial_unapply1; partial_apply1; lambda_eta2] THEN
  MATCH_MP_TAC (prove (`A=B /\ C=D ==> (A,C) = (B,D)`, MESON_TAC[])) THEN
  STRIP_TAC THEN ABS_TAC THEN REWRITE_TAC[partial_apply1; lambda_eta; lambda_eta2] THEN
  ASM_CASES_TAC `(d (x:real^1)):bool` THEN ASM_SIMP_TAC[] THEN CONV_TAC cas_conv THEN
  EXISTS_TAC `(f (x:real^1)):real` THEN SIMP_TAC []
);;

let partial_unapply2_fact = prove (
  `punapp2 (papp1 (f:real^1->real, d)) = (\x. if d x then f (x:real^1) else @v. F), (\y. d (y:real^1))`,
  REWRITE_TAC[partial_unapply2; partial_apply1; lambda_eta2] THEN 
  MATCH_MP_TAC (prove (`A=B /\ C=D ==> (A,C) = (B,D)`, MESON_TAC[])) THEN
  STRIP_TAC THEN ABS_TAC THEN REWRITE_TAC[partial_apply1; lambda_eta; lambda_eta2] THEN ASM_CASES_TAC `(d (x:real^1)):bool` THEN 
  ASM_SIMP_TAC[] THEN CONV_TAC cas_conv THEN
  EXISTS_TAC `(f (x:real^1)):real` THEN SIMP_TAC []
);;


(*
let partial_diff_fact = prove (
  `pdiff (papp1 (f,d)) = 
    papp1 ((\x. @v. ((\x. f (lambda i. x)) diffl v) (x$1)),
      (\x. d x /\ (?v. ((\x.  f (lambda i. x)) diffl v) (x$1))))`
 REWRITE_TAC [partial_diff; partial_unapply1_fact; partial_diff_proto]
 ONCE_REWRITE_TAC [GSYM (REWRITE_CONV [ETA_AX] `\z. (papp1 y z)`)] 
 ABS_TAC THEN ASM_CASES_TAC `x:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1]
 RULE_ASSUM_TAC (MATCH_MP lemma2)
 FIRST_X_ASSUM (X_CHOOSE_TAC `u:real`)
 FIRST_X_ASSUM (fun x -> ASSUME_TAC (CONV_RULE cas_conv x));;
 FIRST_X_ASSUM (fun x -> ALL_TAC)
 ASM_CASES_TAC `(d ((lambda i. a):real^1)):bool` THEN ASM_SIMP_TAC[]
 CONV_TAC cas_conv
*)


(*

`punapp1 (papp1 (f, d)) = ((\y. if d y then f y else &0), \x. d (x:real^1))`

let partial_diff_rule = prove `pdiff (papp1 (f:real^1->real, d:real^1->bool)) = 
 papp1 ( (\x:real^1. @v. ((\x. f (lambda i. x)) diffl v) (x$1)) , (\x:real^1. d x /\ ?v. ((\x. f (lambda i. x)) diffl v) (x$1)) )`;;


REWRITE_TAC [partial_diff; partial_inv; partial_unapply1; partial_diff_proto] THEN
AP_TERM_TAC THEN
MATCH_MP_TAC (prove (`A=B /\ C=D ==> (A,C) = (B,D)`, MESON_TAC[]))
STRIP_TAC THEN ABS_TAC THEN REWRITE_TAC[partial_apply1; lambda_eta; lambda_eta2]
*)

override_interface("+",`padd:(real)option->(real)option->(real)option`);;
override_interface("-",`psub:(real)option->(real)option->(real)option`);;
override_interface("*",`pmul:(real)option->(real)option->(real)option`);;
override_interface("/",`pdiv:(real)option->(real)option->(real)option`);;


let lemma = prove (`!x. x = NONE \/ ?y. x = SOME y`, MESON_TAC[SPEC `\x:(A)option. ((x = NONE) \/ (?y. x = SOME y))` option_INDUCT]);;
let lemma2 = prove (`~(x = NONE) ==> ?y. x = SOME y`, MESON_TAC[lemma]);;

let partial_apply2_beta_right = prove (`papp2 (f, d) (a:(A)option) (SOME b) = 
                                papp1 ((\x. f (lambda i. if i = 1 then x$1 else b)), 
                                       (\x. d (lambda i. if i = 1 then x$1 else b))) a`,
  ASM_CASES_TAC `a:(A)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_apply2] THEN 
  RULE_ASSUM_TAC (MATCH_MP lemma2) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `y:A`) THEN 
  ASM_SIMP_TAC[partial_apply1; partial_apply2; lambda_eta]
);;

let partial_apply2_beta_left = prove (`papp2 (f, d) (SOME a) (b:(A)option) = 
                                papp1 ((\x. f (lambda i. if i = 1 then a else x$1)), 
                                       (\x. d (lambda i. if i = 1 then a else x$1))) b`,
  ASM_CASES_TAC `b:(A)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_apply2] THEN 
  RULE_ASSUM_TAC (MATCH_MP lemma2) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `y:A`) THEN 
  ASM_SIMP_TAC[partial_apply1; partial_apply2; lambda_eta]
);;

let papp1_papp1 = prove (`papp1 (f1, d1) (papp1 (f2, d2) (x:(A)option)) = 
                         papp1 ((\x. f1 (lambda i.(f2 x))), (\x. d2 x /\ d1 (lambda i.(f2 x)))) x`, 
  ASM_CASES_TAC `x:(A)option = NONE` THEN ASM_SIMP_TAC[partial_apply1] THEN RULE_ASSUM_TAC (MATCH_MP lemma2) THEN 
  FIRST_X_ASSUM (X_CHOOSE_TAC `y:A`) THEN ASM_SIMP_TAC[partial_apply1] THEN COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply1]
);;
let papp1_papp2 = prove (`papp1 (f1, d1) (papp2 (f2, d2) (x:(A)option) (y:(A)option)) = 
                          papp2 ((\x:A^2. f1 (lambda i.(f2 x))), (\x:A^2. (d2 x) /\ (d1 (lambda i.(f2 x))))) 
                          (x:(A)option) (y:(A)option)`, 
  ASM_CASES_TAC `x:(A)option = NONE` THEN ASM_CASES_TAC `y:(A)option = NONE` THEN 
  ASM_SIMP_TAC [partial_apply1; partial_apply2] THEN RULE_ASSUM_TAC (MATCH_MP lemma2) THEN 
  FIRST_X_ASSUM (X_CHOOSE_TAC `a:A`) THEN ASM_SIMP_TAC [] THEN FIRST_X_ASSUM (fun x -> ALL_TAC) THEN 
  FIRST_X_ASSUM (X_CHOOSE_TAC `b:A`) THEN ASM_SIMP_TAC[partial_apply2; partial_apply1] THEN 
  COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply1]
);;

let lambda_2_1 = prove (`(lambda i. if i = 1 then x else y)$1 = x`, REWRITE_TAC[lambda_eta]);;
let lambda_2_2 = prove (`((lambda i. if i = 1 then x else y):A^2)$2 = y`, SIMP_TAC [CART_EQ;DIMINDEX_2;ARITH;LAMBDA_BETA]);;

let papp2_papp1_r = prove (
    `papp2 (f1, d1) (a:(A)option) (papp1 (f2, d2) (b:(A)option)) = 
     papp2 ((\x:A^2. f1 ((lambda i.if i = 1 then (x$1) else (f2 ((lambda i.x$2):A^1))):A^2)), 
     (\x:A^2. (d2 ((lambda i. (x$2)):A^1)) /\ (d1 ((lambda i.if i = 1 then (x$1) else (f2 (lambda i. x$2))):A^2)))) 
     (a:(A)option) (b:(A)option)`, 
   ASM_CASES_TAC `a:(A)option = NONE` THEN ASM_CASES_TAC `b:(A)option = NONE` THEN 
   ASM_SIMP_TAC [partial_apply1; partial_apply2] THEN RULE_ASSUM_TAC (MATCH_MP lemma2) THEN 
   FIRST_X_ASSUM (X_CHOOSE_TAC `d:A`) THEN ASM_SIMP_TAC [] THEN FIRST_X_ASSUM (fun x -> ALL_TAC) THEN 
   FIRST_X_ASSUM (X_CHOOSE_TAC `c:A`) THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2] THEN 
   COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2]
);;

(* Proof is copied from above *)
let papp2_papp1_l = prove (
    `papp2 (f1, d1) (papp1 (f2, d2) (a:(A)option)) (b:(A)option) = 
     papp2 ((\x:A^2. f1 ((lambda i.if i = 1 then (f2 ((lambda i.x$1):A^1)) else (x$2)):A^2)), 
     (\x:A^2. (d2 ((lambda i. (x$1)):A^1)) /\ (d1 ((lambda i.if i = 1 then (f2 (lambda i. x$1)) else (x$2)):A^2)))) 
     (a:(A)option) (b:(A)option)`, 
   ASM_CASES_TAC `a:(A)option = NONE` THEN ASM_CASES_TAC `b:(A)option = NONE` THEN 
   ASM_SIMP_TAC [partial_apply1; partial_apply2] THEN RULE_ASSUM_TAC (MATCH_MP lemma2) THEN 
   FIRST_X_ASSUM (X_CHOOSE_TAC `d:A`) THEN ASM_SIMP_TAC [] THEN FIRST_X_ASSUM (fun x -> ALL_TAC) THEN 
   FIRST_X_ASSUM (X_CHOOSE_TAC `c:A`) THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2] THEN 
   COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2]
);;

List.iter cas_eq_thm [
  partial_def; 
  partial_apply1; papp1_papp1; partial_num; partial_inv; 
  partial_apply2; partial_apply2_beta_left; partial_apply2_beta_right; 
  partial_binop; partial_add; partial_sub; partial_mul; partial_div; 
  lambda_eta; lambda_2_1; lambda_2_2; 
  papp1_papp2; papp2_papp1_r; papp2_papp1_l
];;

let dif1 = REWRITE_CONV [partial_diff; partial_inv; partial_def; partial_unapply1_fact; partial_diff_proto; lambda_eta; partial_add; partial_binop; partial_num; partial_apply2_beta_left] `pdiff (\x. pnum 3 + x)` in
let dif1 = REWRITE_RULE [ETA_AX] dif1 in
let dif1 = REWRITE_RULE [partial_unapply1_fact; partial_diff_proto] dif1 in
CONV_RULE cas_conv dif1;;

REWRITE_CONV [partial_diff; partial_inv; partial_def; partial_unapply1_fact; partial_diff_proto; lambda_eta] `pdiff pinv`;;

cas_conv `pnum 3 + pnum 5 + x + (pinv (pnum 4)) * pnum 3 + pinv (y + pnum 1) - pnum 1`;;

(* `((\x. if ~(x = &0) then inv x else @v. F) diffl (-- &1)) (&1)` *)

(*
prioritize_real ();;

let partial_def = new_definition `!(f:A->A) (d:A->bool). partial(f, d) = (f, d)`;;
let partial_of_real_def = new_definition `partial_of_real (f:real->real) = partial (f, \(x:real). T)`;;
let partial_inv = new_definition `pinv = partial (inv, \z. ~(z = &0))`;;
let partial_apply = define `(pappo ((f:A->B), (d:A->bool)) NONE = NONE) /\ pappo (f, d) (SOME x) = if d x then (SOME (f x)) else NONE`;;
let partial_pair = define `(ppair (SOME (x:A)) NONE = NONE) /\ (ppair NONE (SOME (y:B)) = NONE) /\
       (ppair (SOME (x:A)) (SOME (y:B)) = (SOME (x,y)))`;;
let option_of_real_def = new_definition `option_of_real (x:real) = (SOME x)`;;
let option_inv = new_definition `oinv x = pappo pinv x`;;

let option_binop = new_definition `obinop (b:A->B->C) x y = pappo ((\(x,y). b x y), (\(x,y). T)) (ppair x y)`;;
let option_add = new_definition `oadd = obinop real_add`;;

List.iter cas_eq_thm [partial_apply; partial_def; option_inv; option_add; option_binop; option_of_real_def; partial_inv; partial_pair];;
*)

(*
let pi = define `(pinv (SOME x) = if ~(x = &0) then SOME (inv x) else NONE) /\ (pinv NONE = NONE)`;;
let pa = define `padd (SOME x) (SOME y) = (SOME (x + y)) /\ padd NONE v = NONE /\ padd v NONE = NONE`;;
let ps = define `psub (SOME x) (SOME y) = (SOME (x - y)) /\ psub NONE v = NONE /\ psub v NONE = NONE`;;
let pm = define `pmul (SOME x) (SOME y) = (SOME (x * y)) /\ pmul NONE v = NONE /\ pmul v NONE = NONE`;;
let pn = define `(pneg (SOME (x:real)) = SOME (--x)) /\ (pneg NONE = NONE)`;;
let pd = new_definition `pdiv x y = pmul x (pinv y)`;;

let paf = new_definition `paddf f g = \(x:(complex)option). padd (f x) (g x)`;;
let psf = new_definition `psubf f g = \(x:(complex)option). psub (f x) (g x)`;;

List.iter cas_eq_thm [pi; pa; ps; pm; pd; pn];;

cas_conv `(\x. padd (pinv x) (pinv x)) (SOME (Cx(&1)))`;;
cas_conv `(\x. padd (pinv x) (pinv (padd x (SOME (Cx(&1)))))) (SOME(Cx(&1)))` ;;

cas_conv `(\x. padd (pinv x) (pinv x)) (SOME (&1))`;;
cas_conv `(\x. padd (pinv x) (pinv (padd x (SOME (&1))))) (SOME(&1))` ;;

let pu = new_definition `punpart (f:(A)option->(B)option) = \x. @y. (f (SOME x) = (SOME y))`;;
let pv = new_definition `ppart (f:A->(B)option) = \x. if x = NONE then NONE else f (@y. x = SOME y)`;;
let pp = new_definition `pdiff f = ppart (\x. if (f (SOME x) = NONE) then NONE else SOME (diff (punpart f) x))`;;

List.iter cas_eq_thm [pp; pu; pv];;

let option1 = prove (`(NONE = SOME (x:A)) <=> F`, REWRITE_TAC[distinctness "option"]);;

let o1 = prove (`(if P then (SOME x) else NONE) = NONE <=> ~P`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"]);;
let o2 = prove (`(if P then NONE else (SOME x)) = NONE <=> P`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"]);;
let o3 = prove (`(if P then (SOME x) else NONE) = SOME y <=> P /\ x = y`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"; injectivity "option"]);;
let o4 = prove (`(if P then NONE else (SOME x)) = SOME y <=> ~P /\ x = y`, ASM_CASES_TAC `P:bool` THEN ASM_SIMP_TAC[distinctness "option"; injectivity "option"]);;

List.iter cas_eq_thm [o1; o2; o3; o4];;
*)
