let cas_cache = Hashtbl.create 1000;;

let cas_cache_print () =
  Hashtbl.iter (fun a -> function
    | None -> print_term a; print_newline ()
    | Some b -> print_thm b; print_newline ())
;;

let cas_cache_clear () =
  Hashtbl.clear cas_cache
;;


let assumptions = ref [];;
let assumptions_type = ref [];;

let assume_def = new_definition `assume (p:bool) = p`;;
let assume_type_def = new_definition `assume_type (p) = T`;;
let assuming_def = new_definition `assuming x (b:bool) = x`;;
parse_as_infix ("assuming", (1, "left"));;

let assume_conv term =
  let l, r = dest_comb term in
  if l = `assume` then begin
    if List.mem r !assumptions then () 
      else (assumptions := r :: !assumptions; cas_cache_clear ());
    REWRITE_CONV [assume_def] term
  end else if fst (dest_const l) = "assume_type" then begin
    if List.mem r !assumptions_type then () 
      else (assumptions_type := r :: !assumptions_type; cas_cache_clear ());
    REWRITE_CONV [assume_type_def] term
  end else failwith "not an assume term"
;;

let assuming_conv sub_simplify term =
  let opl, r = dest_comb term in
  let op, l = dest_comb opl in
  if fst (dest_const op) <> "assuming" then failwith "not an assuming term" else
  assumptions := r :: !assumptions;
  cas_cache_clear ();
  let ret = sub_simplify l in
  assumptions := List.tl !assumptions;
  cas_cache_clear ();
(*  print_thm ret; print_newline ();*)
  REWRITE_CONV [ret; assuming_def] term
;;

let get_assumed_variables () =
  (List.flatten (map frees !assumptions)) @ (List.flatten (map frees !assumptions_type))
;;

let boolassume_conv term = SIMP_CONV (List.map ASSUME !assumptions) term;;

(* A version that doesn't warn on stdoutput *)
let rec term_of_preterm_nowarn ptm =
(*  stvs_translated := false;*)
  match ptm with
    Varp(s,pty) -> mk_var(s,type_of_pretype pty)
  | Constp(s,pty) -> mk_mconst(s,type_of_pretype pty)
  | Combp(l,r) -> mk_comb(term_of_preterm_nowarn l,term_of_preterm_nowarn r)
  | Absp(v,bod) -> mk_gabs(term_of_preterm_nowarn v,term_of_preterm_nowarn bod)
  | Typing(ptm,pty) -> term_of_preterm_nowarn ptm
;;

let cas_simple_command = ref (function n -> failwith "No such command");;
let cas_applied_command = ref (function n -> function r -> failwith "No such command");;

let add_cas_simple_command name fnctn =
  let old = !cas_simple_command in
  cas_simple_command := fun x -> if x <> name then old x else
  (print_endline ("Command: " ^ name); fnctn ())
;;

let add_cas_applied_command name fnctn =
  let old = !cas_applied_command in
  cas_applied_command := fun x rest -> if x <> name then old x rest else begin
    print_string ("Command: " ^ name ^ " Argument: ");
    print_term (term_of_preterm_nowarn (retypecheck [] rest));
    print_newline ();
    fnctn rest;
  end
;;

let assumptions () =
  List.iter (fun x -> print_qterm x; print_char ' ') (List.rev !assumptions); print_newline ()
in add_cas_simple_command "assumptions" assumptions;;

let about rest =
  let pe = map (fun v -> let s,t = dest_var v in (s,pretype_of_type t)) (get_assumed_variables ()) in
  let tocheck = term_of_preterm_nowarn (retypecheck pe rest) in
  let vars = frees tocheck in
  let isvar l1 l2 =
    try List.iter (fun e1 -> List.iter (fun e2 -> if e1 = e2 then failwith "exists" else ()) l2) l1; false
    with Failure "exists" -> true
  in
  let folder acc assum =
    if isvar vars (frees assum) then assum :: acc else acc
  in
  let matching = List.fold_left folder [] !assumptions in
  List.iter (fun x -> print_qterm x; print_char ' ') matching; print_newline ()
in add_cas_applied_command "about" about;;

let cas_command_handler = function
  | Varp (name, _) -> !cas_simple_command name
  | Combp (Varp (name, _), rest) -> !cas_applied_command name rest
  | _ -> failwith "No such command"
;;
