(* Does only fixed expressions *)
let dint = new_definition `dint (a,b) f = @k. defint (a,b) f k`;;

let dint_cos = prove (`!a b. a <= b ==> dint (a,b) cos = sin b - sin a`, ASM_MESON_TAC[dint; FTC1; DIFF_SIN; DINT_UNIQ]);;
let dint_cos = ONCE_REWRITE_RULE [REWR_CONV (GSYM ETA_AX) `cos`] dint_cos;;
let dint_exp = prove (`!a b. a <= b ==> dint (a,b) exp = exp b - exp a`, ASM_MESON_TAC[dint; FTC1; DIFF_EXP; DINT_UNIQ]);;
let dint_exp = ONCE_REWRITE_RULE [REWR_CONV (GSYM ETA_AX) `exp`] dint_exp;;
let dint_sin_diff = REWRITE_RULE [REAL_MUL_RID; REAL_NEG_NEG] (DIFF_CONV `\x. --(cos x)`);;
let dint_sin = prove (`!a b. a <= b ==> dint (a,b) sin = (--cos b) - (--cos a)`, ASM_MESON_TAC[dint; FTC1; dint_sin_diff; DINT_UNIQ]);;
let dint_sin = ONCE_REWRITE_RULE [REWR_CONV (GSYM ETA_AX) `sin`] dint_sin;;
let dint_pow_diff = 
  SIMP_RULE [REAL_MUL_LINV; REAL_OF_NUM_EQ; NOT_SUC; REAL_MUL_LID; ADD1] (
  ONCE_REWRITE_RULE [REAL_MUL_ASSOC] (REWRITE_RULE [REAL_MUL_AC] (
  SIMP_RULE [REAL_MUL_LINV; REAL_OF_NUM_EQ; NOT_SUC; REAL_MUL_LID] (
  ONCE_REWRITE_RULE [REAL_MUL_ASSOC] (ONCE_REWRITE_RULE[REAL_MUL_SYM] (
  REWRITE_RULE [real_div; REAL_INV_MUL; REAL_MUL_AC] (
  REWRITE_RULE [ADD_SUB; REAL_MUL_LZERO; REAL_MUL_RID; REAL_SUB_RZERO; REAL_POW_2; GSYM ADD1; REAL_OF_NUM_EQ; NOT_SUC] 
  (DIFF_CONV `\x. x pow (n + 1) / (&(n + 1))`))))))))
;;
let dint_pow = prove (`!a b n. a <= b ==> dint (a,b) (\x. (x pow n)) = 
  (\x. inv (&(n + 1)) * x pow (n + 1)) b - (\x. inv (&(n + 1)) * x pow (n + 1)) a`, 
  REWRITE_TAC[dint] THEN ASM_MESON_TAC[SPECL [`\x. inv(&(n + 1)) * x pow (n + 1)`; `\x. x pow n`] FTC1; dint_pow_diff; DINT_UNIQ])
;;

(* Already in multivariate integration there are checkers to write dint_sum, dint_mul, ...  *)

let rec int_conv term =
  let range, fnctn = dest_binop `dint` term in
  if not (is_abs fnctn) then 
    (* As in DIFF_CONV, we first uneta *)
    let th0 = REWR_CONV (GSYM ETA_AX) fnctn in
    let th1 = ONCE_REWRITE_CONV [th0] term in
    let th2 = int_conv (get_res th1) in
    TRANS th1 th2
  else
    let lo, right = dest_comb range in
    let _, left = dest_comb lo in
    let range_thm = REAL_ARITH (mk_binop `(<=)` left right) in
    SIMP_CONV [range_thm; dint_cos; dint_sin; dint_exp; dint_exp; dint_pow] term
;;

