let cas_inout_no = ref 1;;
let cas_outputs = ref [];;
let cas_inputs = ref [];;
let cas_thms = ref [];;

(* preterm_of_term cannot be retypechecked... *)
let rec preterm_of_term_type = function
  | Var(name, ty) -> Typing (Varp(name, Ptycon("",[])), pretype_of_type ty)
  | Const(name, ty) -> Typing (Varp(name, Ptycon("",[])), pretype_of_type ty)
  | Comb(l,r) -> Combp (preterm_of_term_type l, preterm_of_term_type r)
  | Abs(l,r) -> Absp (preterm_of_term_type l, preterm_of_term_type r)
;;

let store input output thm =
  let input = Typing (preterm_of_term_type input, pretype_of_type (type_of input)) in
  cas_inputs := input :: !cas_inputs;
  let output = Typing (preterm_of_term_type output, pretype_of_type (type_of output)) in
  cas_outputs := output :: !cas_outputs;
  cas_thms := thm :: !cas_thms;
  incr cas_inout_no
;;

let rec replace_history preterm =
  match preterm with
  | Combp(a, b) -> Combp(replace_history a, replace_history b)
  | Absp(a, b) -> Absp(replace_history a, replace_history b)
  | Varp(a, t) -> 
    begin try
      if String.length a > 2 && String.sub a 0 2 = "In" then
        let no = int_of_string (String.sub a 2 (String.length a - 2)) in
        List.nth (List.rev !cas_inputs) (no - 1)
      else if String.length a > 3 && String.sub a 0 3 = "Out" then
        let no = int_of_string (String.sub a 3 (String.length a - 3)) in
        List.nth (List.rev !cas_outputs) (no - 1)
      else Varp(a, t)
    with _ -> Varp(a, t) end
  | Typing(a, t) -> Typing(replace_history a, t)
  | Constp(a, t) -> Constp(a, t)
;;
  
let rec replace_last line =
  try 
    let position = String.index line '%' in
    replace_last
       ((String.sub line 0 position) ^ "In" ^ (string_of_int (!cas_inout_no - 1)) ^ 
        (String.sub line (position + 1) (String.length line - position - 1)))
  with
    Not_found -> line
;;

let rec cas_mainloop conv =
  (* For non-ld systems: print_string "> "; flush stdout; *)
  Printf.printf "> %!";
  let line = read_line () in
  if line = "exit" then () else
  if line = "" then cas_mainloop conv else
  if (String.length line > 1 && line.[0] = '/' && line.[1] = '/') then (print_endline line; cas_mainloop conv) else
  let line = replace_last line in
  let preterm =
    try
      let preterm, l = (parse_preterm o lex o explode) line in
      if l <> [] then raise Exit else
      preterm
    with _ -> failwith "Unable to parse preterm"
  in
  let preterm = replace_history preterm in
  try cas_command_handler preterm; cas_mainloop conv with Failure "No such command" ->
  try
    let pe = map (fun v -> let s,t = dest_var v in (s,pretype_of_type t)) (get_assumed_variables ()) in
    let preterm = try retypecheck pe preterm with _ -> failwith "Unable to assign types" in
    let input = 
      try term_of_preterm preterm 
      with _ -> failwith "Unable to assign types to term"
    in
    (*print_type (type_of input); print_newline ();*)
    print_string "In"; print_int !cas_inout_no;
    print_string " := "; print_term input;
    print_newline (); flush stdout;
    let thm = conv input in
    let output = get_res thm in
    print_string "Out"; print_int !cas_inout_no;
    print_string " := "; print_term output;
    print_newline (); flush stdout;
    store input output thm;
    cas_mainloop conv
  with 
  | End_of_file -> ()
  | Noparse -> 
    print_endline ("Incorrect input: " ^ line);
    print_endline "Error: Noparse exception";
    cas_mainloop conv
  | Failure x ->
    print_endline ("Incorrect input: " ^ line);
    print_endline ("Error: Failure: " ^ x);
    cas_mainloop conv
;;

let cas_print_thms () =
  let iterator t =
    open_hbox ();
    print_thm t; 
    close_box (); 
    print_newline ()
  in List.iter iterator (List.rev !cas_thms)
in add_cas_simple_command "theorems" cas_print_thms;;
