let BINOP2_CONV conv1 conv2 tm =
  let lop,r = dest_comb tm in
  let op,l = dest_comb lop in
  MK_COMB(AP_TERM op (conv1 l),conv2 r)
;;


let conv_or_not_conv conv term =
  try conv term
  with _ ->
    try
      let thm = REWRITE_RULE[NOT_CLAUSES] (MK_COMB ((REFL `(~)`), (conv (mk_comb (`(~)`, term))))) in
      REWRITE_CONV[thm] term
    with _ -> REFL term
;;


let cas_int_conv term =
  let cas_real_conv term =
    try_conv_list [REAL_RAT_REDUCE_CONV; REAL_POLY_CONV;
      REWRITE_CONV[GSYM REAL_INV_MUL] ] term
  in
  try
    let int_thm_lst =
      [int_eq; int_le; int_lt; int_ge; int_gt;
       int_of_num_th; int_neg_th; int_add_th; int_mul_th;
       int_sub_th; int_abs_th; int_max_th; int_min_th; int_pow_th]
    in
    let thm = REWRITE_CONV int_thm_lst (mk_comb (`dest_int`, term)) in
    let thm = TRANS thm (cas_real_conv (get_res thm)) in
    INT_OF_REAL_THM thm
  with Failure _ -> REFL term
;;

let lemma = prove(`!n. ii pow (2 * n) = --Cx(&1) pow n`, INDUCT_TAC THENL [
  NUM_REDUCE_TAC THEN REWRITE_TAC[complex_pow];
  ASM_REWRITE_TAC [MULT_SUC; COMPLEX_POW_ADD; COMPLEX_POW_II_2; complex_pow]
]);;

let lemma2 = prove (`!n. ii pow (2 * n + 1) = ii * --Cx(&1) pow n`,
  REWRITE_TAC[GSYM ADD1; complex_pow; lemma]
);;

(* Calculates `ii pow n` when n is a num const, less than linear. *)
let COMPLEX_II_POW_CONV term =
  try let l, p = dest_binop `complex_pow` (get_res (REWRITE_CONV [SYM ii] term)) in
    if l <> `ii` then failwith "not ii" else
    let n = dest_numeral p in
    let zero, one, two = Num.num_of_int 0, Num.num_of_int 1, Num.num_of_int 2 in
    if n </ two then failwith "Only from 2" else
    let quo, modu = quo_num n two, mod_num n two in
    let thm = SPEC (mk_numeral quo) (if modu = zero then lemma else lemma2) in
    let conv = NUM_REDUCE_CONV (concl thm) in
    REWRITE_RULE[ii] (EQ_MP conv thm)
  with _ -> REFL term
;;

let complex_unreal = prove(`!a b. complex (a,b) = ii * Cx b + Cx a`,
  REWRITE_TAC[ii] THEN SIMPLE_COMPLEX_ARITH_TAC
);;
let complex_mulout = prove( `complex(a,b) * complex(c,d) = complex(a*c-b*d,a*d+b*c)`, 
  SIMPLE_COMPLEX_ARITH_TAC
);;
let complex_addout = prove( `complex(a,b) + complex(c,d) = complex(a+c,b+d)`, 
  SIMPLE_COMPLEX_ARITH_TAC
);;
let complex_poly_conv term =
 (REWRITE_CONV[complex_unreal]
  THENC COMPLEX_POLY_CONV
  THENC REWRITE_CONV[GSYM complex_unreal; CX_DEF; ii]
  THENC COMPLEX_POLY_CONV  
  THENC REWRITE_CONV[complex_mulout;complex_addout]
  THENC REAL_RAT_REDUCE_CONV) term
;;

(* SIMPLE_COMPLEX_ARITH_TAC eg for `Cx(&2 * y) = Cx(&2) * Cx(y)` *)
let cas_simple_complex_conv =
  let real_unreal = GSYM CX_DEF in
  let imag_unreal = prove (`complex (&0, x) = ii * Cx(x)`,
    REWRITE_TAC[ii] THEN SIMPLE_COMPLEX_ARITH_TAC) in
  fun term ->
    let thm = REWRITE_CONV [COMPLEX_EQ; RE; IM; CX_DEF; complex_add; complex_neg;
                         complex_sub; complex_mul; complex_inv; complex_div; ii] term in
    let thm = TRANS thm (REAL_RAT_REDUCE_CONV (get_res thm)) in
    let thm = TRANS thm (my_depth_conv REAL_POLY_CONV (get_res thm)) in
    (*print_thm thm; print_newline ();*)
    TRANS thm (REWRITE_CONV [COMPLEX; real_unreal; SYM ii; imag_unreal; complex_unreal] (get_res thm))
;;

(*let diff = new_definition `diff f = \x. @v. (f diffl v) (x)`;;*)
let diff = new_definition `diff f = \x. if (?v. (f diffl v) x) then (@v. (f diffl v) x) else &0`;;

let diff_thm = prove (`(!x. (f diffl (g x)) x) ==> diff f = \x. g x`,
  STRIP_TAC THEN REWRITE_TAC[diff] THEN ABS_TAC THEN
  SUBGOAL_THEN `?v. (f diffl v) x` ASSUME_TAC THENL [ASM_MESON_TAC[]; ALL_TAC] THEN
  ASM_SIMP_TAC[] THEN  MATCH_MP_TAC SELECT_UNIQUE THEN BETA_TAC THEN ASM_MESON_TAC[DIFF_UNIQ])
;;

let diff_impl_thm = prove (`
       (!x. P x ==> (f diffl (g x)) x) 
   /\  (!x. ~(P x) ==> ~(?v. (f diffl v) x)) 
   ==> diff f = \x. if (P x) then (g x) else (&0)`,
  STRIP_TAC THEN REWRITE_TAC[diff] THEN ABS_TAC THEN
  ASM_CASES_TAC `(P (x:real)):bool` THEN ASM_SIMP_TAC[] THEN
  ASSUME_TAC (MP (SPEC `x:real` (ASSUME `!x. P x ==> (f diffl g x) x`)) (ASSUME `(P (x:real)):bool`)) THEN
  SUBGOAL_THEN `?v. (f diffl v) x` ASSUME_TAC THENL [ASM_MESON_TAC[]; ASM_SIMP_TAC[]] THEN
  MATCH_MP_TAC SELECT_UNIQUE THEN BETA_TAC THEN ASM_MESON_TAC[DIFF_UNIQ])
;;

let diff_conv term =
  let op, fnctn = dest_comb term in
  if op <> `diff` then failwith "not a diff term" else
  let d_thm = DIFF_CONV fnctn in
  let var, app = dest_abs (snd (dest_comb (concl d_thm))) in
  let diffll, res = dest_binop `(diffl)` (fst (dest_comb app)) in
  let res_fnctn = mk_abs (var, res) in
  let d_thm2 = prove (mk_comb (mk_binop `(diffl)` diffll (mk_comb (res_fnctn, var)), var), REWRITE_TAC[ETA_AX; d_thm]) in
  MATCH_MP diff_thm (GEN var d_thm2)
;;


let magic1 = new_axiom `!x. x = &0 ==> ~(?v. (inv diffl v) x)`;;
(*
let magic2 = new_axiom `(!x. P x ==> ~(?v. (f diffl v) x)) /\ (!x. (g diffl h x) x) ==> (!x. P x ==> ~(?v. ((f + g) diffl v) x))`;;
let magic3 = new_axiom `(!x. P x ==> ~(?v. (f diffl v) x)) ==> (!x. P (x + a) ==> ~(?v. ((\x. f (x + a)) diffl v) (x + a))`;;
*)

(*
  `!x. x = &0 ==> ~(?v. ((\x. inv x) diffl v) x))`
   STRIP_TAC THEN STRIP_TAC
   ASM_SIMP_TAC[] THEN
   STRIP_TAC THEN
   FIRST_X_ASSUM (fun x -> MP_TAC (MATCH_MP DIFF_CONT x))
FIRST_X_ASSUM (fun _ -> ALL_TAC)
*)

let diff_impl_prepare_conv term =
  let op, fnctn = dest_comb term in
  if op <> `diff` then failwith "not a diff term" else
  let d_thm = DIFF_CONV fnctn in
  let var, app = dest_abs (snd (dest_comb (concl d_thm))) in
  let prereq, app = dest_imp app in
  let diffll, res = dest_binop `(diffl)` (fst (dest_comb app)) in
  let res_fnctn = mk_abs (var, res) in
  let diff_thm = SPECL [diffll; mk_abs (var, prereq); res_fnctn] (GEN_ALL diff_impl_thm) in
  REWRITE_RULE [d_thm] diff_thm
;;

let diff_impl_conv term =
  MATCH_MP (diff_impl_prepare_conv term) magic1
;;



let sum_conv term =
  let range, fnctn = dest_binop `sum` term in
  let _, cnt = dest_comb range in
  REWRITE_CONV[num_CONV cnt; sum] term
;;

(* Works in general, but slow, good for sides *)
let binom_slow_conv term =
  let arg = rand term in
  let bottom = rand arg in
  let top = rand (rator arg) in
  let t1 = try num_CONV top with _ -> REFL top
  and t2 = try num_CONV bottom with _ -> REFL bottom in
  REWRITE_CONV [t1; t2; binom] term
;;

let binom_lemma = prove (`!n k. 0 < FACT n * FACT k`, REPEAT INDUCT_TAC THEN ASM_SIMP_TAC[FACT; ARITH; LT_MULT; LT_0; FACT_LT]);;

let binom_conv term =
  let arg = rand term in
  let bottom = rand arg in
  let top = rand (rator arg) in
  if dest_numeral top </ dest_numeral bottom then binom_slow_conv term else
  let topn = mk_numeral (dest_numeral top -/ dest_numeral bottom) in
  let ftopn = mk_comb (`FACT`, topn) in
  let fbottom = mk_comb (`FACT`, bottom) in
  let ftop = mk_comb (`FACT`, top) in
  let fmul = mk_binop `(*):num->num->num` ftopn fbottom in (**)
  let ath = REWRITE_RULE [ADD_0; MULT_AC] (SPECL [ftop; fmul; term; `0`] DIVMOD_UNIQ) in
  let thm = SYM(SPECL [topn; bottom] BINOM_FACT) in
  let thm' = REWRITE_RULE[ARITH; MULT_AC] (CONJ thm (SPECL [bottom; topn] binom_lemma)) in
  SYM (CONJUNCT1 (MP ath thm'))
;;

let prime_lemma = prove(`!p. prime(p) <=> ~(p = 1) /\ !x. (x < p + 1 \/ p = 0) ==> num_divides x p ==> x = 1 \/ x = p`,
  REWRITE_TAC [prime] THEN GEN_TAC THEN EQ_TAC THEN STRIP_TAC THEN CONJ_TAC THEN ASM_SIMP_TAC[] 
  THEN GEN_TAC THEN STRIP_TAC THEN FIRST_ASSUM (fun x -> ASSUME_TAC (REWRITE_RULE [GSYM LT_SUC_LE; ADD1] (MATCH_MP DIVIDES_LE x))) 
  THEN ASM_MESON_TAC []
);;

(* Slow, but works... *)
let prime_conv term =
  let t1 = REWRITE_CONV [prime; ARITH; prime_lemma] term in
  let t2 = EXPAND_CASES_CONV (get_res t1) in
  TRANS t1 t2
;;
  

let decimal_conv term =
  let lt, rt = dest_binop `real_div` term in
  let rt = snd (dest_comb rt) in
  let rval = dest_numeral rt in
  let rec div num =
    if num >/ (Int 1) then
      if Num.mod_num num (Int 10) = (Int 0) then div (num // (Int 10))
      else false
    else num = (Int 1)
  in
  if div rval then REWRITE_CONV [GSYM DECIMAL] term
  else failwith "Not a quotient"
;;

let abs_thm = prove (`!x. &0 <= x ==> abs x = x`, SIMP_TAC [GSYM REAL_ABS_REFL]);;
