prioritize_real ();;

let partial_def = new_definition `!(f:A^N->A) (d:A^N->bool). partial(f, d) = (f, d)`;;

let partial_apply1 = define
  `(papp1 ((f:A^1->A), (d:A^1->bool)) NONE = NONE) /\
   (papp1 (f, d) (SOME x) = if (d (lambda i.x)) then (SOME (f (lambda i.x))) else NONE)`;;

let partial_apply2 = define
  `(papp2 ((f:A^2->A), (d:A^2->bool)) (SOME x) (SOME y) =
     if (d (lambda i.if i = 1 then x else y)) then
       (SOME (f (lambda i.if i = 1 then x else y))) else NONE) /\
   (papp2 (f, d) NONE v = NONE) /\
   (papp2 (f, d) v NONE = NONE)`;;

let partial_binop = new_definition
  `pbinop (f:A->A->A) x y = papp2 ((\x:A^2. (f:A->A->A) (x$1) (x$2)),(\x:A^2.T)) x y`;;
let partial_add = new_definition `padd = pbinop real_add`;;
let partial_sub = new_definition `psub = pbinop real_sub`;;
let partial_mul = new_definition `pmul = pbinop real_mul`;;
let partial_inv = new_definition
  `pinv = papp1 (partial ((\x:real^1. inv (x$1)), \x:real^1. ~((x$1) = &0)))`;;
let partial_neg = new_definition `pneg = papp1 ((\x:real^1. --(x$1)), \x. T)`;;
let partial_sqrt = new_definition `psqrt = papp1 ((\x. sqrt (x$1)), (\x. (x$1) >= &0))`;;
let partial_div = new_definition `pdiv x y = pmul x (pinv y)`;;
let partial_num = new_definition `pnum n = SOME (&n)`;;

let partial_diff_proto = new_definition `pdiff_proto (f:real^1->real, d:real^1->bool) =
  ( (\x:real^1. if d x /\
    ?v. ((\x. f (lambda i. x)) diffl v) (x$1) then @v. ((\x. f (lambda i. x)) diffl v)
  (x$1) else &0) , (\x:real^1. d x /\ ?v. ((\x. f (lambda i. x)) diffl v) (x$1)) )`
;;

let partial_unapply = new_definition
  `punapp f = ((\x:real^1. @v:real. (f(SOME (x$1))) = (SOME v)),
               (\x:real^1. ?v. (f (SOME (x$1))) = (SOME v)))`
;;

let partial_diff = new_definition `pdiff f = papp1 (pdiff_proto (punapp f))`;;

let lambda_eta = REWRITE_RULE [ARITH_RULE `1 <= 1`; DIMINDEX_GE_1] (SPEC `1` LAMBDA_BETA);;
let lambda_eta2 = prove (`(lambda i. x$1) = (x:A^1)`,
  SIMP_TAC [CART_EQ;DIMINDEX_1;ARITH;LAMBDA_BETA;ARITH_RULE `1 <= i /\ i <= 1 <=> i = 1`]
);;

let SELECT_LEMMA = GEN_REWRITE_RULE (BINDER_CONV o LAND_CONV o ONCE_DEPTH_CONV) [EQ_SYM_EQ] SELECT_REFL;;

let option_case = prove (`~(x = NONE) ==> ?y. x = SOME y`,
  MESON_TAC[SPEC `\x:(A)option. ((x = NONE) \/ (?y. x = SOME y))` option_INDUCT]
);;

let partial_unapply_fact = prove (
  `punapp (papp1 (f:real^1->real, d)) = (\x. if d x then f (x:real^1) else @v. F), (\y. d (y:real^1))`,
  REWRITE_TAC[partial_unapply; partial_apply1; lambda_eta2] THEN
  MATCH_MP_TAC (prove (`A=B /\ C=D ==> (A,C) = (B,D)`, MESON_TAC[])) THEN
  STRIP_TAC THEN ABS_TAC THEN REWRITE_TAC[partial_apply1; lambda_eta; lambda_eta2] THEN
  ASM_CASES_TAC `(d (x:real^1)):bool` THEN ASM_SIMP_TAC[] THEN
  REWRITE_TAC[distinctness "option"; injectivity "option"; SELECT_LEMMA] THEN
  EXISTS_TAC `(f (x:real^1)):real` THEN SIMP_TAC []
);;

let papp_in_domain = prove(
  `papp1 ((\x. if d x then y x else &0), (\x. d x)) = papp1 ((\x. y x), (\x. d x))`,
  GEN_REWRITE_TAC (ONCE_DEPTH_CONV o BINOP_CONV) [GSYM ETA_AX] THEN ABS_TAC THEN
  ASM_CASES_TAC `x:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1] THEN
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `u:real`) THEN
  ASM_SIMP_TAC[partial_apply1]
);;

(*
  `pdiff (papp1 (f,d)) = papp1 ((\x. @v. ((\x. f (lambda i. x)) diffl v) (x$1)),
      (\x. d x /\ (?v. ((\x.  f (lambda i. x)) diffl v) (x$1))))`
  Not Provable
*)

let partial_apply2_beta_right = prove (`papp2 (f, d) (a:(A)option) (SOME b) =
                                papp1 ((\x. f (lambda i. if i = 1 then x$1 else b)),
                                       (\x. d (lambda i. if i = 1 then x$1 else b))) a`,
  ASM_CASES_TAC `a:(A)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_apply2] THEN
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `y:A`) THEN
  ASM_SIMP_TAC[partial_apply1; partial_apply2; lambda_eta]
);;

let partial_apply2_beta_left = prove (`papp2 (f, d) (SOME a) (b:(A)option) =
                                papp1 ((\x. f (lambda i. if i = 1 then a else x$1)),
                                       (\x. d (lambda i. if i = 1 then a else x$1))) b`,
  ASM_CASES_TAC `b:(A)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_apply2] THEN
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `y:A`) THEN
  ASM_SIMP_TAC[partial_apply1; partial_apply2; lambda_eta]
);;

let papp1_papp1 = prove (`papp1 (f1, d1) (papp1 (f2, d2) (x:(A)option)) =
                         papp1 ((\x. f1 (lambda i.(f2 x))), (\x. d2 x /\ d1 (lambda i.(f2 x)))) x`,
  ASM_CASES_TAC `x:(A)option = NONE` THEN ASM_SIMP_TAC[partial_apply1] THEN RULE_ASSUM_TAC (MATCH_MP option_case) THEN
  FIRST_X_ASSUM (X_CHOOSE_TAC `y:A`) THEN ASM_SIMP_TAC[partial_apply1] THEN COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply1]
);;

let papp1_papp2 = prove (`papp1 (f1, d1) (papp2 (f2, d2) (x:(A)option) (y:(A)option)) =
                          papp2 ((\x:A^2. f1 (lambda i.(f2 x))), (\x:A^2. (d2 x) /\ (d1 (lambda i.(f2 x)))))
                          (x:(A)option) (y:(A)option)`,
  ASM_CASES_TAC `x:(A)option = NONE` THEN ASM_CASES_TAC `y:(A)option = NONE` THEN
  ASM_SIMP_TAC [partial_apply1; partial_apply2] THEN RULE_ASSUM_TAC (MATCH_MP option_case) THEN
  FIRST_X_ASSUM (X_CHOOSE_TAC `a:A`) THEN ASM_SIMP_TAC [] THEN FIRST_X_ASSUM (fun x -> ALL_TAC) THEN
  FIRST_X_ASSUM (X_CHOOSE_TAC `b:A`) THEN ASM_SIMP_TAC[partial_apply2; partial_apply1] THEN
  COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply1]
);;

let lambda_2_1 = prove (`(lambda i. if i = 1 then x else y)$1 = x`, REWRITE_TAC[lambda_eta]);;
let lambda_2_2 = prove (`((lambda i. if i = 1 then x else y):A^2)$2 = y`, SIMP_TAC [CART_EQ;DIMINDEX_2;ARITH;LAMBDA_BETA]);;
let lambda_2_2_2 = prove (`((lambda) f):A^2$2 = f 2`, SIMP_TAC [CART_EQ;DIMINDEX_2;ARITH;LAMBDA_BETA]);;

let papp2_papp1_r = prove (
    `papp2 (f1, d1) (a:(A)option) (papp1 (f2, d2) (b:(A)option)) =
     papp2 ((\x:A^2. f1 ((lambda i.if i = 1 then (x$1) else (f2 ((lambda i.x$2):A^1))):A^2)),
     (\x:A^2. (d2 ((lambda i. (x$2)):A^1)) /\ (d1 ((lambda i.if i = 1 then (x$1) else (f2 (lambda i. x$2))):A^2))))
     (a:(A)option) (b:(A)option)`,
   ASM_CASES_TAC `a:(A)option = NONE` THEN ASM_CASES_TAC `b:(A)option = NONE` THEN
   ASM_SIMP_TAC [partial_apply1; partial_apply2] THEN RULE_ASSUM_TAC (MATCH_MP option_case) THEN
   FIRST_X_ASSUM (X_CHOOSE_TAC `d:A`) THEN ASM_SIMP_TAC [] THEN FIRST_X_ASSUM (fun x -> ALL_TAC) THEN
   FIRST_X_ASSUM (X_CHOOSE_TAC `c:A`) THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2] THEN
   COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2]
);;

(* Proof is nearly the same as above, should integrate *)
let papp2_papp1_l = prove (
    `papp2 (f1, d1) (papp1 (f2, d2) (a:(A)option)) (b:(A)option) =
     papp2 ((\x:A^2. f1 ((lambda i.if i = 1 then (f2 ((lambda i.x$1):A^1)) else (x$2)):A^2)),
     (\x:A^2. (d2 ((lambda i. (x$1)):A^1)) /\ (d1 ((lambda i.if i = 1 then (f2 (lambda i. x$1)) else (x$2)):A^2))))
     (a:(A)option) (b:(A)option)`,
   ASM_CASES_TAC `a:(A)option = NONE` THEN ASM_CASES_TAC `b:(A)option = NONE` THEN
   ASM_SIMP_TAC [partial_apply1; partial_apply2] THEN RULE_ASSUM_TAC (MATCH_MP option_case) THEN
   FIRST_X_ASSUM (X_CHOOSE_TAC `d:A`) THEN ASM_SIMP_TAC [] THEN FIRST_X_ASSUM (fun x -> ALL_TAC) THEN
   FIRST_X_ASSUM (X_CHOOSE_TAC `c:A`) THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2] THEN
   COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply2; partial_apply1; lambda_2_1; lambda_2_2]
);;

let papp2_same = prove (`papp2 (f, d) x x = papp1 ((\x:real^1. f ((lambda i. x$1):real^2)), \x:real^1. d ((lambda i. x$1):real^2)) x`,
  ASM_CASES_TAC `x:(real)option = NONE` THEN
  ASM_SIMP_TAC[partial_apply1; partial_apply2] THEN
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN
  FIRST_X_ASSUM (X_CHOOSE_TAC `y:real`) THEN
  ASM_SIMP_TAC[partial_apply1; partial_apply2; lambda_eta; prove (`(if i = 1 then y:real else y) = y`, MESON_TAC[])]
);;

let diff_but_for_any_point = prove(`!w y. ((~(y = w)) ==> (f diffl (g y)) y ==> ((\x. if ~(x = w) then f x else @v. F) diffl (g y)) y)`,
    GEN_TAC THEN GEN_TAC THEN STRIP_TAC THEN REWRITE_TAC[diffl; LIM] THEN STRIP_TAC THEN GEN_TAC THEN STRIP_TAC THEN
    FIRST_X_ASSUM (fun eps -> FIRST_X_ASSUM (fun dif -> ASSUME_TAC eps THEN ASSUME_TAC (MATCH_MP dif eps)))
    THEN FIRST_X_ASSUM (X_CHOOSE_TAC `dd:real`) THEN EXISTS_TAC `min dd (abs (y - w))` THEN
    REWRITE_TAC [real_min] THEN
    FIRST_X_ASSUM(fun x ->
      ASSUME_TAC (CONJUNCT1 x) THEN ASSUME_TAC (CONJUNCT2 x)) THEN COND_CASES_TAC THEN ASM_SIMP_TAC[]
    THENL [
      GEN_TAC THEN
      FIRST_X_ASSUM(fun dd -> FIRST_X_ASSUM(fun main -> DISCH_TAC THEN FIRST_X_ASSUM(fun newa ->
        ASSUME_TAC (MATCH_MP main newa) THEN ASSUME_TAC dd THEN ASSUME_TAC (CONJUNCT1 newa) THEN
        ASSUME_TAC (CONJUNCT2 newa)))) THEN
      ASM_SIMP_TAC [MP (MP (REAL_FIELD `dd <= abs (y - w) ==> abs (x - &0) < dd ==> ~(y + x = w)`)
        (ASSUME `dd <= abs (y - w)`)) (ASSUME `abs (x - &0) < dd`)]
    ; ALL_TAC] THEN
    ASM_SIMP_TAC [REAL_FIELD `~(y = w) ==> &0 < abs (y - w)`] THEN
    GEN_TAC THEN STRIP_TAC THEN
    SIMP_TAC [MP (REAL_FIELD `abs (x - &0) < abs (y - w) ==> ~(y + x = w)`) (ASSUME `abs (x - &0) < abs (y - w)`)]
    THEN FIRST_ASSUM (fun a6 -> FIRST_X_ASSUM (fun a5 -> FIRST_X_ASSUM (fun a4 ->
      FIRST_X_ASSUM (fun a3 -> MATCH_MP_TAC (SPEC `x:real` a3) THEN ASSUME_TAC a4 THEN
      ASSUME_TAC a5 THEN ASSUME_TAC a6)))) THEN
    ASM_MESON_TAC [REAL_FIELD `abs (x - &0) < abs (y - w) ==> ~(dd <= abs (y - w)) ==> abs (x - &0) < dd`]
);;

let diff_but_for_axis = prove(`!w y. ((y > w) ==> (f diffl (g y)) y ==> ((\x. if (x >= w) then f x else @v. F) diffl (g y)) y)`,
  GEN_TAC THEN GEN_TAC THEN STRIP_TAC THEN REWRITE_TAC[diffl; LIM] THEN STRIP_TAC THEN GEN_TAC THEN STRIP_TAC THEN
  FIRST_X_ASSUM (fun eps -> FIRST_X_ASSUM (fun dif -> ASSUME_TAC eps THEN ASSUME_TAC (MATCH_MP dif eps)))
  THEN FIRST_X_ASSUM (X_CHOOSE_TAC `dd:real`) THEN EXISTS_TAC `min dd (y - w)` THEN REWRITE_TAC [real_min] THEN
  FIRST_X_ASSUM(fun x -> ASSUME_TAC (CONJUNCT1 x) THEN ASSUME_TAC (CONJUNCT2 x)) THEN COND_CASES_TAC THEN ASM_SIMP_TAC[]
  THENL [
    GEN_TAC THEN
    FIRST_X_ASSUM(fun dd -> FIRST_X_ASSUM(fun main -> DISCH_TAC THEN FIRST_X_ASSUM(fun newa ->
      ASSUME_TAC (MATCH_MP main newa) THEN ASSUME_TAC dd THEN ASSUME_TAC (CONJUNCT1 newa) THEN
      ASSUME_TAC (CONJUNCT2 newa)))) THEN
    COND_CASES_TAC THEN COND_CASES_TAC THEN REPEAT (FIRST_X_ASSUM MP_TAC) THEN REAL_ARITH_TAC;
    STRIP_TAC
  ] THENL [REPEAT (FIRST_X_ASSUM MP_TAC) THEN REAL_ARITH_TAC;
    GEN_TAC THEN STRIP_TAC
  ] THEN
  COND_CASES_TAC THEN COND_CASES_TAC THEN REPEAT (FIRST_X_ASSUM MP_TAC) THEN TRY REAL_ARITH_TAC THEN
  STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC THEN STRIP_TAC THEN FIRST_X_ASSUM (fun x -> ASSUME_TAC (SPEC `x:real` x))
  THEN REPEAT (FIRST_X_ASSUM MP_TAC) THEN TRY REAL_ARITH_TAC
);;

let diff_inv_fact = prove (`!y. (~(y = &0) ==>
  ((\x. if ~(x = &0) then inv x else @v. F) diffl (--real_div (&1) (y pow 2))) y)`,
  GEN_TAC THEN STRIP_TAC THEN
  FIRST_ASSUM (fun x -> ASSUME_TAC (MATCH_MP (SPEC `&0:real` diff_but_for_any_point) x)) THEN
  FIRST_X_ASSUM (fun x -> MATCH_MP_TAC x) THEN
  ASM_SIMP_TAC [DIFF_CONV `inv`];
);;

let diff_inv_exists = prove(`~((x:real^1)$1 = &0) /\
  (?v. ((\x. if ~(x = &0) then inv x else @v. F) diffl v) (x$1)) <=> ~(x$1 = &0)`,
  EQ_TAC THEN STRIP_TAC THEN ASM_SIMP_TAC[] THEN EXISTS_TAC `--real_div (&1) (((x:real^1)$1) pow 2)`
  THEN ASM_SIMP_TAC[diff_inv_fact]
);;

let diff_inv_value = prove(
  `(if ~((x:real^1)$1 = &0)
         then @v. ((\x. if ~(x = &0) then inv x else @v. F) diffl v) (x$1)
         else &0) = (if ~(x$1 = &0) then --real_div (&1) (((x:real^1)$1) pow 2) else &0)`,
  ASM_MESON_TAC [SELECT_UNIQUE; diff_inv_fact; DIFF_UNIQ]);;

let back_binop = prove (
  `papp1 ((\x:real^1. binop (a x) (b x)), d) x = (pbinop binop) (papp1((\x:real^1. a x), d) x) (papp1((\x:real^1. b x), d) x)`,
  ASM_CASES_TAC `x:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_apply2; partial_add; partial_binop] THEN
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN  FIRST_X_ASSUM (X_CHOOSE_TAC `z:real`) THEN ASM_SIMP_TAC[partial_apply1] THEN
  COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply2; lambda_eta; lambda_2_2] 
);;


let back_app = prove (
  `papp1 ((\x:real^1. x$1), (\x:real^1. T)) z = z`, 
  ASM_CASES_TAC `z:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1] THEN 
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `v:real`) THEN 
  ASM_SIMP_TAC[partial_apply1; lambda_eta]
);;

let back_noappr = prove (
  `papp1 ((\x:real^1. binop (a x) b), d) x = (pbinop binop) (papp1((\x:real^1. a x), d) x) (SOME b)`,
  ASM_CASES_TAC `x:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_binop; partial_apply2_beta_right] THEN 
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `z:real`) THEN 
  ASM_SIMP_TAC[partial_apply1; lambda_eta; lambda_2_2] THEN
  COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply1; lambda_eta]
);;

let back_noappl = prove (
  `papp1 ((\x:real^1. binop a (b x)), d) x = (pbinop binop) (SOME a) (papp1((\x:real^1. b x), d) x)`,
  ASM_CASES_TAC `x:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1; partial_binop; partial_apply2_beta_left] THEN 
  RULE_ASSUM_TAC (MATCH_MP option_case) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `z:real`) THEN 
  ASM_SIMP_TAC[partial_apply1; lambda_eta; lambda_2_2] THEN
  COND_CASES_TAC THEN ASM_SIMP_TAC[partial_apply1; lambda_eta]
);;

let back = 
  (List.map (fun op -> SPEC op (GEN_ALL back_binop)) [`real_add`; `real_mul`; `real_sub`; `real_div`]) @ 
  (List.map GSYM [partial_add; partial_mul; partial_sub; partial_div]) @
  [back_app; back_noappr; back_noappl; GSYM partial_num]
;;

let full_conv term =
  let t1 = (cas_conv THENC (ONCE_REWRITE_CONV [GSYM ETA_AX])) term in
  let l1 = List.map (fun op -> SPEC op (GEN_ALL back_noappr)) [`real_add`; `real_mul`; `real_sub`] in
  let l2 = List.map (fun op -> SPEC op (GEN_ALL back_noappl)) [`real_add`; `real_mul`; `real_sub`] in
  let t2 = REWRITE_CONV (l1 @ l2) (get_res t1) in
  let t3 = REWRITE_CONV back (get_res t2) in
  TRANS t1 (TRANS t2 t3)
;;


REWRITE_CONV [partial_diff_proto; partial_unapply_fact; partial_diff; partial_sqrt] `pdiff psqrt`;;

let lemma = prove (`!x. x pow 2 >= &0`, 
  ONCE_REWRITE_TAC [GSYM REAL_POW2_ABS] THEN REWRITE_TAC[REAL_POW_2; GSYM REAL_ABS_MUL; real_ge; REAL_ABS_POS]
);;

let partial_sqrt_diff = prove(`x$1 >= &0 ==> (?v. ((\x. if x >= &0 then sqrt x else @v. F) diffl v) ((x:real^1)$1)) = x$1 > &0`,
  STRIP_TAC THEN EQ_TAC THENL [ALL_TAC;
    STRIP_TAC THEN
    FIRST_ASSUM (fun x -> ASSUME_TAC (MATCH_MP (SPECL [`&0`; `(x:real^1)$1`] diff_but_for_axis) x)) THEN
    FIRST_X_ASSUM (fun x -> ASSUME_TAC (GEN `f:real->real` (GEN `g:real->real` x))) THEN
    EXISTS_TAC `(inv (&2 * sqrt  ((x:real^1)$1)):real) * &1` THEN
    FIRST_X_ASSUM (fun x -> MATCH_MP_TAC x) THEN
    MATCH_MP_TAC (DIFF_CONV `sqrt`) THEN
    FIRST_X_ASSUM MP_TAC THEN
    REAL_ARITH_TAC] THEN

ASM_CASES_TAC `(x:real^1)$1 > &0` THEN ASM_SIMP_TAC[] THEN
FIRST_X_ASSUM (fun x -> FIRST_X_ASSUM (fun y -> SIMP_TAC [MP (REAL_ARITH `(x:real^1)$1 >= &0 /\ ~(x$1 > &0) ==> x$1 = &0`) (CONJ x y)]))

 THEN REWRITE_TAC[NOT_EXISTS_THM; diffl; LIM; NOT_FORALL_THM] THEN
GEN_TAC THEN EXISTS_TAC `&1` THEN ASM_SIMP_TAC [REAL_ARITH `&0 < &1`] THEN
REWRITE_TAC[NOT_EXISTS_THM] THEN GEN_TAC THEN
ASM_CASES_TAC `&0 < d` THEN ASM_SIMP_TAC [NOT_FORALL_THM] THEN
REWRITE_TAC [REAL_ADD_LID; REAL_ARITH `&0 >= &0 <=> T`; REAL_ARITH `x - &0 = x`; SQRT_0] THEN 
REWRITE_TAC [real_sub; real_div] THEN
ONCE_REWRITE_TAC [REAL_MUL_SYM; REAL_NEG_MINUS1] THEN
SUBGOAL_THEN `min (d / &2) ((&1 / (&2 + v)) pow 2) >= &0` ASSUME_TAC THENL [
  MATCH_MP_TAC (REAL_ARITH `x > &0 /\ y >= &0 ==> min x y >= &0`) THEN
  ASM_SIMP_TAC [REAL_ARITH `&0 < d ==> d / &2 > &0`; lemma]; ALL_TAC
] THEN
SUBGOAL_THEN `min (d / &2) ((&1 / (&1 + v)) pow 2) >= &0` ASSUME_TAC THENL [
  MATCH_MP_TAC (REAL_ARITH `x > &0 /\ y >= &0 ==> min x y >= &0`) THEN
  ASM_SIMP_TAC [REAL_ARITH `&0 < d ==> d / &2 > &0`; lemma]; ALL_TAC
] THEN

ASM_CASES_TAC `v = -- &1` THENL [
  EXISTS_TAC `min (d / &2) ((&1 / (&2 + v)) pow 2)`;
  EXISTS_TAC `min (d / &2) ((&1 / (&1 + v)) pow 2)`
] THEN MATCH_MP_TAC (TAUT `A /\ ~B ==> ~(A ==> B)`) THEN ASM_SIMP_TAC []
THEN FIRST_X_ASSUM (fun x -> FIRST_X_ASSUM (fun _ -> FIRST_X_ASSUM (fun _ -> ASSUME_TAC x)))

THEN REWRITE_TAC[
  REAL_ARITH `&0 < abs x <=> ~(x = &0)`; 
  REAL_ARITH `((&1 / (&2 + -- &1)) pow 2) = &1`;
  REAL_ARITH `-- &1 * -- &1 = &1`; real_min
] THEN COND_CASES_TAC THEN
  ASM_SIMP_TAC [REAL_ARITH `&0 < d ==> ~(d / &2 = &0)`;
               REAL_ARITH `&0 < d ==> abs (d / &2) < d`;
               REAL_ARITH `&0 < d ==> d / &2 >= &0`;
               REAL_FIELD `x pow 2 = &0 <=> x = &0`; REAL_ARITH `~(&1 = &0)`; SQRT_1;
               REAL_ARITH `~(abs (inv (&1) * &1 + &1) < &1)`;
               REAL_FIELD `~(v = -- &1) ==> ~(&1 / (&1 + v) = &0)`;
               REAL_ARITH `~(d / &2 <= &1) ==> abs (&1) < d`]
THEN TRY (REPEAT CONJ_TAC)
THENL [
 ALL_TAC; ALL_TAC; REWRITE_TAC[REAL_POW_2] THEN REWRITE_TAC[REAL_ABS_MUL; GSYM REAL_POW_2; REAL_POW2_ABS]
  THEN ASM_SIMP_TAC[REAL_FIELD `&0 < d ==> ~(d / &2 <= (&1 / (&1 + v)) pow 2) ==> (&1 / (&1 + v)) pow 2 < d`];
  ALL_TAC]
THEN MATCH_MP_TAC (REAL_ARITH `!x y. x >= &1 - y ==> ~(abs (x + y) < &1)`)
THEN REWRITE_TAC [POW_2_SQRT_ABS]
THENL [
  REWRITE_TAC[real_ge; REAL_ARITH `&1 - &1 = &0`] THEN
  MATCH_MP_TAC REAL_LE_MUL THEN 
  ASM_SIMP_TAC[MATCH_MP SQRT_POS_LE (MP (REAL_ARITH `&0 < d ==> &0 <= d / &2`) (ASSUME `&0 < d`))] THEN
  MATCH_MP_TAC REAL_LE_INV THEN
  ASM_SIMP_TAC[REAL_ARITH `&0 < d ==> &0 <= d / &2`]; ALL_TAC; ALL_TAC
] THEN
REWRITE_TAC[REAL_FIELD `inv ((&1 / (&1 + v)) pow 2) = (&1 + v) pow 2`; REAL_ARITH `&1 / x = inv x`] THEN
let l1 = prove (`~(w = &0) ==> w pow 2 * abs (inv w) = abs(w)`, 
  REWRITE_TAC[real_abs] THEN ASM_CASES_TAC `&0 <= w` THEN ASM_SIMP_TAC[REAL_LE_INV_EQ; REAL_LT_INV_EQ] THEN
  REWRITE_TAC [REAL_POW_2; REAL_MUL_AC] THEN ASM_SIMP_TAC[REAL_MUL_RNEG; REAL_MUL_RINV] THEN REAL_ARITH_TAC
) in
ASM_SIMP_TAC[REWRITE_RULE [REAL_ARITH `~(&1 + v = &0) <=> ~(v = -- &1)`] (SPEC `&1 + v` (GEN_ALL l1))] THEN
TRY REAL_ARITH_TAC THEN
REWRITE_TAC[real_sub; real_ge; REAL_NEG_NEG; REAL_MUL_LNEG; REAL_MUL_LID] THEN

let l2 = prove(`&0 < d ==> d <= (inv w) pow 2 ==> w <= inv d * sqrt d`,
  REPEAT STRIP_TAC THEN
  ONCE_REWRITE_TAC[MATCH_MP (GSYM POW_2_SQRT) (MATCH_MP (ONCE_REWRITE_RULE [GSYM REAL_LE_INV_EQ] (REAL_ARITH `&0 < d ==> &0 <= d`)) (ASSUME `&0 < d`))] THEN
  REWRITE_TAC [GSYM (MATCH_MP SQRT_MUL (CONJ (SPEC `2` (MATCH_MP REAL_POW_LE (MP (ONCE_REWRITE_RULE [GSYM REAL_LE_INV_EQ] (REAL_FIELD `&0 < d ==> &0 <= d`)) (ASSUME `&0 < d`)))) (MP (REAL_FIELD `&0 < d ==> &0 <= d`) (ASSUME `&0 < d`))))] THEN
  MATCH_MP_TAC REAL_LE_RSQRT THEN
  ASM_SIMP_TAC[REAL_FIELD `&0 < d ==> inv d pow 2 * d = inv d`] THEN 
  GEN_REWRITE_TAC LAND_CONV [GSYM REAL_INV_INV] THEN
  MATCH_MP_TAC REAL_LE_INV2 THEN
  REWRITE_TAC [GSYM REAL_POW_INV] THEN ASM_SIMP_TAC[]
) in
FIRST_X_ASSUM (fun x -> ASM_SIMP_TAC [MATCH_MP (MATCH_MP l2 (ONCE_REWRITE_RULE [REAL_ARITH `&0 < d <=> &0 < d / &2`] (ASSUME `&0 < d`))) (REWRITE_RULE [REAL_ARITH `&1 / x = inv x`] x)])

);;

overload_interface("+",`padd:(real)option->(real)option->(real)option`);;
overload_interface("-",`psub:(real)option->(real)option->(real)option`);;
overload_interface("*",`pmul:(real)option->(real)option->(real)option`);;
overload_interface("/",`pdiv:(real)option->(real)option->(real)option`);;
overload_interface("--",`pneg:(real)option->(real)option`);;
overload_interface("&",`pnum:num->(real)option`);;

let prioritize_option() = prioritize_overload(`:((real)option)`);;

let partial_theorems = [
  partial_def;
  partial_apply1; papp1_papp1; partial_num; partial_inv;
  partial_apply2; partial_apply2_beta_left; partial_apply2_beta_right;
  partial_binop; partial_add; partial_sub; partial_mul; partial_div; partial_neg;
  lambda_eta; lambda_2_1; lambda_2_2; lambda_2_2_2;
  papp1_papp2; papp2_papp1_r; papp2_papp1_l; papp2_same;
  partial_unapply_fact;
  partial_diff; partial_diff_proto;
  diff_inv_value; diff_inv_exists; papp_in_domain
];;

let diff_thm = prove (`(!x. ((\x:real. f x) diffl (g x)) x) ==> 
  papp1((\(x:real^1). @v. ((\x:real. f x) diffl v) (x$1)), (\(x:real^1). ?v. ((\x:real. f x) diffl v) (x$1))) = 
  papp1((\(x:real^1). g (x$1)), \(x:real^1). T)`,
  STRIP_TAC THEN AP_TERM_TAC THEN MATCH_MP_TAC (prove (`A=B /\ C=D ==> (A,C) = (B,D)`, MESON_TAC[])) THEN
  STRIP_TAC THEN ABS_TAC THENL [
    ASM_MESON_TAC[SELECT_UNIQUE; DIFF_UNIQ];
    EQ_TAC THEN SIMP_TAC[] THEN EXISTS_TAC `(g ((x:real^1)$1)):real` THEN ASM_MESON_TAC []
  ]
);;


List.iter cas_eq_thm partial_theorems;; 

cas_conv `pnum 3 + pnum 5 + x + (pinv (pnum 4)) * pnum 3 + pinv (y + pnum 1) - pnum 1`;;
cas_conv `pdiff pinv`;;



let diff_conv term =
  let fst_thm = cas_conv term in
  let term = get_res fst_thm in
  let op, fnctn = dest_comb term in
  if op <> `papp1:(real^1->real)#(real^1->bool)->(real)option->(real)option` then failwith "not a papp term" else
  let p1 = fst (dest_pair fnctn) in
  let hilbert = snd (dest_abs p1) in
  let diff = snd(dest_abs(snd(dest_comb hilbert))) in
  let the_fnctn = fst (dest_binop `(diffl)` (fst (dest_comb diff))) in
  let the_thm = MATCH_MP diff_thm (DIFF_CONV the_fnctn) in
  TRANS fst_thm the_thm
;;

let diff_conv2 term =
  let fst = diff_conv term in
  let snd = full_conv (get_res fst) in
  TRANS fst snd
;;

diff_conv2 `pdiff (\x. pnum 3 + x)`;;
diff_conv2 `pdiff (\x. pnum 3 + x * pnum 4)`;;
diff_conv2 `pdiff (\y. pnum 3 + y * y + (pinv (pnum 4)) * pnum 3 + (pnum 2) * (y + pnum 1) - pnum 1)`;;
cas_conv `pnum 3 + pnum 5 + y + (pinv (pnum 4)) * pnum 3 + pinv (y + pnum 1) - pnum 1`;;

let diff_thm2 = prove (`
  (!x. ~(x = w) ==> (f diffl (g x)) x) ==> 
  papp1((\(x:real^1). @v. ((\x:real. if ~(x = w) then f x else @v. F) diffl v) (x$1)), 
        (\(x:real^1). (~(x$1 = w) /\ d (x$1)) /\ ?v. ((\x:real. if ~(x = w) then f x else @v. F) diffl v) (x$1))) = 
  papp1((\(x:real^1). g (x$1)), \(x:real^1). ~(x$1 = w) /\ d (x$1))`,
  STRIP_TAC THEN
  let t = MATCH_MP (TAUT `(a ==> b ==> c) ==> (a ==> b) ==> (a ==> c)`) (SPECL [`w:real`; `x:real`] diff_but_for_any_point) in
  let t2 = MATCH_MP (MESON [] `(!x. A x ==> B x) ==> (!x. A x) ==> (!x. B x)`) (GEN `x:real` t) in
  FIRST_X_ASSUM (fun x -> ASSUME_TAC (MATCH_MP t2 x)) THEN
  GEN_REWRITE_TAC (ONCE_DEPTH_CONV o BINOP_CONV) [GSYM ETA_AX] THEN ABS_TAC THEN
  ASM_CASES_TAC `x:(real)option = NONE` THEN ASM_SIMP_TAC[partial_apply1] THEN
  FIRST_X_ASSUM (fun x -> ASSUME_TAC (MATCH_MP option_case x)) THEN FIRST_X_ASSUM (X_CHOOSE_TAC `u:real`) THEN
  ASM_SIMP_TAC[partial_apply1; lambda_eta]  THEN 
  ASM_CASES_TAC `~(u:real = w)` THEN ASM_SIMP_TAC[] THEN
  ASM_CASES_TAC `(d (u:real)):bool` THEN ASM_SIMP_TAC[] THEN
  FIRST_X_ASSUM (fun x -> ALL_TAC) THEN
  FIRST_X_ASSUM (fun notuw -> FIRST_X_ASSUM (fun _ -> FIRST_X_ASSUM (fun main -> ASSUME_TAC (MATCH_MP main notuw)))) THEN
  COND_CASES_TAC THENL [
    AP_TERM_TAC THEN ASM_MESON_TAC[DIFF_UNIQ; SELECT_UNIQUE]; ASM_MESON_TAC[]
  ]
);;

let diff_conv3 term =
  let fst_thm = cas_conv term in
  let term = get_res fst_thm in
  let op, fnctn = dest_comb term in
  if op <> `papp1:(real^1->real)#(real^1->bool)->(real)option->(real)option` then failwith "not a papp term" else
  let p1 = fst (dest_pair fnctn) in
  let hilbert = snd (dest_abs p1) in
  let diff = snd(dest_abs(snd(dest_comb hilbert))) in
  let the_fnctn = fst (dest_binop `(diffl)` (fst (dest_comb diff))) in
  let variab, fnctn_term = dest_abs the_fnctn in
  let to_diff = snd (dest_comb (fst (dest_comb fnctn_term))) in
  let a_thm = DIFF_CONV (mk_abs (variab, to_diff)) in
  fst_thm, a_thm
;;

let diff_conv4 term =
  let t1, t2 = diff_conv3 term in
  let t1 = REWRITE_RULE [ETA_AX] t1 in
  let t2 = REWRITE_RULE [ETA_AX] (SPEC `\(x:real). T` (GEN `d:real->bool` (MATCH_MP diff_thm2 t2))) in
  TRANS t1 t2
;;

diff_conv3 `pdiff (\x. pnum 4 / x + pnum 1)`;;

(*diff_conv `pdiff (pnum 3 + pnum 5 + y + (pinv (pnum 4)) * pnum 3 + pinv (y + pnum 1) - pnum 1)`;;*)

(*cas_conv `pdiff (\y. pnum 3 + pnum 5 + y + (pinv (pnum 4)) * pnum 3 + pinv (y + pnum 1) - pnum 1)`;;*)

(REWRITE_CONV [partial_diff_proto; partial_unapply_fact; partial_diff; partial_sqrt] THENC cas_conv THENC REWRITE_CONV [MATCH_MP (TAUT `(A ==> B = C) ==> (A /\ B <=> A /\ C)`) partial_sqrt_diff]) `pdiff psqrt`;;

(* SPECL [`sqrt`; `\x. inv (&2 * sqrt x) * &1`] (GEN_ALL (SPECL [`&0`; `x:real`] diff_but_for_axis));;*)

let rec convert term =
(*  print_term term;
  print_newline ();*)
  if is_var term then mk_var (fst (dest_var term), `:((real)option)`) else
  if not (is_comb term) then mk_comb (`SOME:real->(real)option`,term) else
  let (l, r) = dest_comb term in
  if l = `real_of_num` then mk_comb (`pnum`, r) else
  if l = `real_neg` then mk_comb (`pneg`, convert r) else
  if l = `real_inv` then mk_comb (`pinv`, convert r) else
  let (m, l) = dest_comb l in
  if m = `real_add` then mk_binop `padd` (convert l) (convert r) else
  if m = `real_mul` then mk_binop `pmul` (convert l) (convert r) else
  if m = `real_sub` then mk_binop `psub` (convert l) (convert r) else
  if m = `real_div` then mk_binop `pdiv` (convert l) (convert r) else
  let name, _ = dest_var l in
  mk_var (name, `:((real)option)`)
;;


let diff_conv5 term =
  let thm1 = (diff_conv4 THENC cas_conv THENC REWRITE_CONV [REAL_POW_2]) term in
  let term = get_res thm1 in
  let term, dom = dest_comb (snd (dest_comb term)) in
  let _, fn = dest_comb term in
  let var, subfn = dest_abs fn in
  let newsubfun = convert subfn in
  let varn, vart = dest_var var in
  let newvar = mk_var (varn, `:((real)option)`) in
(*  let newsubfunm = newsubfun + pnum 0 * x*)
  let newfun = mk_abs (newvar, newsubfun) in
  let thm2 = GSYM ((cas_conv THENC REWRITE_CONV [SIMP_RULE [ARITH] (SPECL [`x:real`; `2`] REAL_POW_EQ_0); REAL_POW_2]) newfun) in
   TRANS thm1 thm2
;;

let diff_conv6 = diff_conv5 THENC REWRITE_CONV [GSYM partial_div];;
  
diff_conv6 `pdiff (\x. &3 + &4 * x + pinv x)`;;
diff_conv6 `pdiff (\x.  &7 * x * x + &2 / x + x * &y)`;;
diff_conv6 `pdiff (\x. (SOME pi) * x * x + &2 / x + x * &y)`;;
