(* A version of sprint_term, modified to return the string and not print it
   on standard output. Modified from version 2.0. *)

(* ------------------------------------------------------------------------- *)
(* Printer for types.                                                        *)
(* ------------------------------------------------------------------------- *)

let sprint_term =
  let rec sprint_term prec tm =
(*    try string_of_num(dest_complex_const tm) with _ -> *)
    try string_of_num(dest_numeral tm) with Failure _ ->
    try let tms = dest_list tm in ("[" ^ (sprint_term_sequence "; " 0 tms) ^ "]")
    with Failure _ ->
    let hop,args = strip_comb tm in
    if is_abs hop & args = [] then sprint_binder prec hop else
    let s0 = NAME_OF hop
    and ty0 = type_of hop in
    let s = try if not !reverse_interface_mapping then fail()
                else fst(find (fun (s,(s',ty)) -> s' = s0 &
                                             can (type_match ty ty0) [])
                         (!the_interface))
            with Failure _ -> s0 in
    try if s = "EMPTY" & args = [] then "{}" else fail()
    with Failure _ ->
    try if s <> "INSERT" then fail() else
        let mems,oth = splitlist (dest_binary "INSERT") tm in
        if is_const oth & fst(dest_const oth) = "EMPTY" then
          ("{" ^ (sprint_term_sequence ", " 14 mems) ^ "}")
        else fail()
    with Failure _ ->
    try if not (s = "GSPEC") then fail() else
        let bod = snd(strip_exists(body(rand tm))) in
        let bod1,fabs = dest_comb bod in
        let bod2,babs = dest_comb bod1 in
        let c = rator bod2 in
        if fst(dest_const c) <> "SETSPEC" then fail() else
        ("{" ^ (sprint_term 0 fabs) ^ " | " ^ (sprint_term 0 babs) ^ "}")
    with Failure _ ->
    try if not (s = "GABS") then fail() else
        let vs,bod = dest_gabs tm in
         (
          (if prec = 0 then "" else "(") ^
          "\\" ^
          (sprint_term 999 vs) ^
          ". " ^
          (sprint_term 0 bod) ^
          (if prec = 0 then "" else ")")
         )
    with Failure _ ->
    try let eqs,bod = dest_let tm in
        (
         (if prec = 0 then "" else "(") ^
         "let " ^
         (sprint_term 0 (mk_eq(hd eqs))) ^
         (String.concat "" 
            (List.map (fun (v,t) -> "and " ^ (sprint_term 0 (mk_eq(v,t)))) (tl eqs))
         ) ^
         "in " ^
         (sprint_term 0 bod) ^
         (if prec = 0 then "" else ")")
        )
    with Failure _ -> try
      if s <> "DECIMAL" then fail() else
      let n_num = dest_numeral (hd args)
      and n_den = dest_numeral (hd(tl args)) in
      let s_num = string_of_num(quo_num n_num n_den) in
      let s_den = implode(tl(explode(string_of_num
                      (n_den +/ (mod_num n_num n_den))))) in
      ("#"^s_num^(if n_den = Int 1 then "" else ".")^s_den)
    with Failure _ ->
    if s = "COND" & List.length args = 3 then
      (
       (if prec = 0 then "" else "(") ^
       "if " ^
       (sprint_term 0 (hd args)) ^ 
       " then " ^
       (sprint_term 0 (hd(tl args))) ^
       " else " ^
       (sprint_term 0 (hd(tl(tl args)))) ^
       (if prec = 0 then "" else ")")
      )
    else if is_prefix s & List.length args = 1 then
      (
       (if prec = 1000 then "(" else "") ^
       s ^
       (if isalnum s or
         s = "--" &
         (try let l,r = dest_comb(hd args) in
              mem (fst(dest_const l))
                  ["real_of_num"; "int_of_num";
                   "real_neg"; "int_neg"]
          with Failure _ -> false) or
         s = "~" & List.length args = 1 & is_neg(hd args)
        then " " else ""
       ) ^
       (sprint_term 999 (hd args)) ^
       (if prec = 1000 then ")" else "")
      )
    else if IS_BINDER s & List.length args = 1 & is_gabs (hd args) then
      sprint_binder prec tm
    else if IS_INFIX s & List.length args = 2 then
      let bargs =
        if ARIGHT s then
          let tms,tmt = splitlist (DEST_BINARY hop) tm in tms@[tmt]
        else
          let tmt,tms = rev_splitlist (DEST_BINARY hop) tm in tmt::tms in
      let newprec = FIXITY s in
      (
       (if newprec <= prec then "(" else "") ^
       (sprint_term newprec (hd bargs)) ^
       (String.concat "" (List.map 
         (fun x -> (if s = "," then ""
                    else " ") ^
                   s ^
                   " " ^
                   (sprint_term newprec x))
         (tl bargs))
       ) ^
       (if newprec <= prec then ")" else "")
      )
    else if (is_const hop or is_var hop) & args = [] then
      let s' = if IS_BINDER s or IS_INFIX s or IS_PREFIX s
               then "("^s^")" else s in
      s'
    else
      let l,r = dest_comb tm in
      (
       (if prec = 1000 then "(" else "") ^
       (sprint_term 999 l) ^
       (if try mem (fst(dest_const l)) ["real_of_num"; "int_of_num"]
           with Failure _ -> false
        then "" else " ") ^
       (sprint_term 1000 r) ^
       (if prec = 1000 then ")" else "")
      )
      
  and sprint_term_sequence sep prec tms =
    if tms = [] then "" else
    (
     (sprint_term prec (hd tms)) ^
     let ttms = tl tms in
     if ttms = [] then ""
     else (sep ^ (sprint_term_sequence sep prec ttms))
    )

  and sprint_binder prec tm =
    let s,(vs,bod) =
      if is_gabs tm then "\\",strip_gabs tm else
      let st,abs = dest_comb tm in
      let s = NAME_OF st in
      s,splitlist (dest_binder_vorc s) tm in
    (
     (if prec = 0 then "" else "(") ^
     s ^
     (if isalnum s then " " else "") ^
     (String.concat "" (List.map (fun x -> (sprint_term 0 x) ^ " ") (butlast vs))) ^
     (sprint_term 0 (last vs)) ^
     ". " ^
     (sprint_term 0 bod) ^
     (if prec = 0 then "" else ")")
    )
  in sprint_term 0
;;

